from __future__ import annotations
from typing import Type, List, Union, TYPE_CHECKING

import copy
import numpy as np
import matplotlib.pyplot as plt

from ...qaoa_components import Hamiltonian
if TYPE_CHECKING:
    from ...optimizers.logger_vqa import Logger
from ...utilities import (
    qaoa_probabilities,
    bitstring_energy,
    convert2serialize,
    delete_keys_from_dict,
)
from ...backends.basebackend import QAOABaseBackend, QAOABaseBackendStatevector
from ...backends.qaoa_analytical_sim import QAOABackendAnalyticalSimulator


def most_probable_bitstring(cost_hamiltonian, measurement_outcomes):
    """
    Computing the most probable bitstring
    """
    mea_out = list(measurement_outcomes.values())
    index_likliest_states = np.argwhere(mea_out == np.max(mea_out))
    # degeneracy = len(index_likliest_states)
    solutions_bitstrings = [
        list(measurement_outcomes.keys())[e[0]] for e in index_likliest_states
    ]

    return {
        "solutions_bitstrings": solutions_bitstrings,
        "bitstring_energy": bitstring_energy(cost_hamiltonian, solutions_bitstrings[0]),
    }


class QAOAResult:
    """
    A class to handle the results of QAOA workflows

    Parameters
    ----------
    log: `Logger`
        The raw logger generated from the training vqa part of the QAOA.
    method: `str`
        Stores the name of the optimisation used by the classical optimiser
    cost_hamiltonian: `Hamiltonian`
        The cost Hamiltonian for the problem statement
    type_backend: `QAOABaseBackend`
        The type of backend used for the experiment
    """

    def __init__(
        self,
        log: Logger,
        method: Type[str],
        cost_hamiltonian: Type[Hamiltonian],
        type_backend: Type[QAOABaseBackend],
    ):
        """
        init method
        """
        self.__type_backend = type_backend
        self.method = method
        self.cost_hamiltonian = cost_hamiltonian

        self.evals = {
            "number_of_evals": log.func_evals.best[0],
            "jac_evals": log.jac_func_evals.best[0],
            "qfim_evals": log.qfim_func_evals.best[0],
        }

        self.intermediate = {
            "angles": np.array(log.param_log.history).tolist(),
            "cost": log.cost.history,
            "measurement_outcomes": log.measurement_outcomes.history,
            "job_id": log.job_ids.history,
        }

        self.optimized = {
            "angles": np.array(log.param_log.best[0]).tolist()
            if log.param_log.best != []
            else [],
            "cost": log.cost.best[0] if log.cost.best != [] else None,
            "measurement_outcomes": log.measurement_outcomes.best[0]
            if log.measurement_outcomes.best != []
            else {},
            "job_id": log.job_ids.best[0] if len(log.job_ids.best) != 0 else [],
            "eval_number": log.eval_number.best[0]
            if len(log.eval_number.best) != 0
            else [],
        }

        self.most_probable_states = (
            most_probable_bitstring(
                cost_hamiltonian, self.get_counts(log.measurement_outcomes.best[0])
            )
            if type_backend != QAOABackendAnalyticalSimulator
            and log.measurement_outcomes.best != []
            else []
        )
        # def __repr__(self):
        #     """Return an overview over the parameters and hyperparameters
        #     Todo
        #     ----
        #     Split this into ``__repr__`` and ``__str__`` with a more verbose
        #     output in ``__repr__``.
        #     """
        #     string = "Optimization Results:\n"
        #     string += "\tThe solution is " + str(self.solution['degeneracy']) + " degenerate" "\n"
        #     string += "\tThe most probable bitstrings are: " + str(self.solution['bitstring']) + "\n"
        #     string += "\tThe associated cost is: " + str(self.optimized['cost']) + "\n"

        #     return (string)

        # if we are using a shot adaptive optimizer, we need to add the number of shots to the result
        if log.n_shots.history != []:
            self.n_shots = log.n_shots.history

    def asdict(
        self,
        keep_cost_hamiltonian: bool = True,
        complex_to_string: bool = False,
        intermediate_measurements: bool = True,
        exclude_keys: List[str] = [],
    ):
        """
        Returns a dictionary with the results of the optimization, where the dictionary is serializable.
        If the backend is a statevector backend, the measurement outcomes will be the statevector,
        meaning that it is a list of complex numbers, which is not serializable.
        If that is the case, and complex_to_string is true the complex numbers are converted to strings.

        Parameters
        ----------
        keep_cost_hamiltonian: `bool`
            If True, the cost hamiltonian is kept in the dictionary. If False, it is removed.
        complex_to_string: `bool`
            If True, the complex numbers are converted to strings. If False, they are kept as complex numbers.
            This is useful for the JSON serialization.
        intermediate_measurements: bool, optional
            If True, intermediate measurements are included in the dump.
            If False, intermediate measurements are not included in the dump.
            Default is True.
        exclude_keys: `list[str]`
            A list of keys to exclude from the returned dictionary.

        Returns
        -------
        return_dict: `dict`
            A dictionary with the results of the optimization, where the dictionary is serializable.
        """

        return_dict = {}
        return_dict["method"] = self.method
        if keep_cost_hamiltonian:
            return_dict["cost_hamiltonian"] = convert2serialize(self.cost_hamiltonian)
        return_dict["evals"] = self.evals
        return_dict["most_probable_states"] = self.most_probable_states

        complex_to_str = (
            lambda x: str(x)
            if isinstance(x, np.complex128) or isinstance(x, complex)
            else x
        )

        # if the backend is a statevector backend, the measurement outcomes will be the statevector,
        # meaning that it is a list of complex numbers, which is not serializable.
        # If that is the case, and complex_to_string is true the complex numbers are converted to strings.
        if complex_to_string:
            return_dict["intermediate"] = {}
            for key, value in self.intermediate.items():
                # Measurements and Cost may require casting
                if "measurement" in key:
                    if len(value) > 0:
                        if intermediate_measurements is False:
                            # if intermediate_measurements is false, the intermediate measurements are not included
                            return_dict["intermediate"][key] = []
                        elif isinstance(
                            value[0], np.ndarray
                        ):  # Statevector -> convert complex to str
                            return_dict["intermediate"][key] = [
                                [complex_to_str(item) for item in list_]
                                for list_ in value
                                if (
                                    isinstance(list_, list)
                                    or isinstance(list_, np.ndarray)
                                )
                            ]
                        else:  # All other case -> cast numpy into
                            return_dict["intermediate"][key] = [
                                {k_: int(v_) for k_, v_ in v.items()} for v in value
                            ]
                    else:
                        pass
                elif "cost" == key and (
                    isinstance(value[0], np.float64) or isinstance(value[0], np.float32)
                ):
                    return_dict["intermediate"][key] = [float(item) for item in value]
                else:
                    return_dict["intermediate"][key] = value

            return_dict["optimized"] = {}
            for key, value in self.optimized.items():
                # If wavefunction do complex to str
                if "measurement" in key and (
                    isinstance(value, list) or isinstance(value, np.ndarray)
                ):
                    return_dict["optimized"][key] = [
                        complex_to_str(item) for item in value
                    ]
                # if dictionary, convert measurement values to integers
                elif "measurement" in key and (isinstance(value, dict)):
                    return_dict["optimized"][key] = {
                        k: int(v) for k, v in value.items()
                    }
                else:
                    return_dict["optimized"][key] = value

                if "cost" in key and (
                    isinstance(value, np.float64) or isinstance(value, np.float32)
                ):
                    return_dict["optimized"][key] = float(value)
        else:
            return_dict["intermediate"] = self.intermediate
            return_dict["optimized"] = self.optimized

        # if we are using a shot adaptive optimizer, we need to add the number of shots to the result,
        # so if attribute n_shots is not empty, it is added to the dictionary
        if getattr(self, "n_shots", None) is not None:
            return_dict["n_shots"] = self.n_shots

        return (
            return_dict
            if exclude_keys == []
            else delete_keys_from_dict(return_dict, exclude_keys)
        )

    @classmethod
    def from_dict(
        cls, dictionary: dict, cost_hamiltonian: Union[Hamiltonian, None] = None
    ):
        """
        Creates a Results object from a dictionary (which is the output of the asdict method)
        Parameters
        ----------
        dictionary: `dict`
            The dictionary to create the QAOA Result object from
        Returns
        -------
        `Result`
            The Result object created from the dictionary
        """

        # deepcopy the dictionary, so that the original dictionary is not changed
        dictionary = copy.deepcopy(dictionary)

        # create a new instance of the class
        result = cls.__new__(cls)

        # set the attributes of the new instance, using the dictionary
        for key, value in dictionary.items():
            setattr(result, key, value)

        # if there is an input cost hamiltonian, it is added to the result
        if cost_hamiltonian is not None:
            result.cost_hamiltonian = cost_hamiltonian

        # if the measurement_outcomes are strings, they are converted to complex numbers
        if not isinstance(
            result.optimized["measurement_outcomes"], dict
        ) and isinstance(result.optimized["measurement_outcomes"][0], str):
            for i in range(len(result.optimized["measurement_outcomes"])):
                result.optimized["measurement_outcomes"][i] = complex(
                    result.optimized["measurement_outcomes"][i]
                )

            for i in range(len(result.intermediate["measurement_outcomes"])):
                for j in range(len(result.intermediate["measurement_outcomes"][i])):
                    result.intermediate["measurement_outcomes"][i][j] = complex(
                        result.intermediate["measurement_outcomes"][i][j]
                    )

        # if the measurement_outcomes are complex numbers, the backend is set to QAOABaseBackendStatevector
        if not isinstance(
            result.optimized["measurement_outcomes"], dict
        ) and isinstance(result.optimized["measurement_outcomes"][0], complex):
            setattr(result, "_QAOAResult__type_backend", QAOABaseBackendStatevector)
        else:
            setattr(result, "_QAOAResult__type_backend", "")

        # return the object
        return result

    @staticmethod
    def get_counts(measurement_outcomes):
        """
        Converts probabilities to counts when the measurement outcomes are a numpy array,
        that is a state vector

        Parameters
        ----------
        measurement_outcomes: `Union[np.array, dict]`
            The measurement outcome as returned by the Logger.
            It can either be a statevector or a count dictionary

        Returns
        -------
        `dict`
            The count dictionary obtained either throught the statevector or
            the actual measurement counts.
        """

        if isinstance(measurement_outcomes, type(np.array([]))):
            measurement_outcomes = qaoa_probabilities(measurement_outcomes)

        return measurement_outcomes

    def plot_cost(
        self, figsize=(10, 8), label="Cost", linestyle="--", color="b", ax=None
    ):
        """
        A simpler helper function to plot the cost associated to a QAOA workflow

        Parameters
        ----------
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        label: `str`
            The label of the cost line, defaults to 'Cost'.
        linestyle: `str`
            The linestyle of the poloit. Defaults to '--'.
        color: `str`
            The color of the line. Defaults to 'b'.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax.plot(
            range(
                1,
                self.evals["number_of_evals"]
                - self.evals["jac_evals"]
                - self.evals["qfim_evals"]
                + 1,
            ),
            self.intermediate["cost"],
            label=label,
            linestyle=linestyle,
            color=color,
        )

        ax.set_ylabel("Cost")
        ax.set_xlabel("Number of function evaluations")
        ax.legend()
        ax.set_title("Cost history")

        return

    def plot_probabilities(
        self,
        n_states_to_keep=None,
        figsize=(10, 8),
        label="Probability distribution",
        color="tab:blue",
        ax=None,
    ):
        """
        Helper function to plot the probabilities corresponding to each basis states
        (with prob != 0) obtained from the optimized result

        Parameters
        ----------
        n_states_to_keep: 'int
            If the user passes a value, the plot will compile with the given value of states.
            Else,  an upper bound will be calculated depending on the
            total size of the measurement outcomes.
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        label: `str`
            The label of the cost line, defaults to 'Probability distribution'.
        color: `str`
            The color of the line. Defaults to 'tab:blue'.
        ax: 'matplotlib.axes._subplots.AxesSubplot'
            Axis on which to plot the graph. Deafults to None
        """

        outcome = self.optimized["measurement_outcomes"]

        # converting to counts dictionary if outcome is statevector
        if isinstance(outcome, type(np.array([]))):
            outcome = self.get_counts(outcome)
            # setting norm to 1 since it might differ slightly for statevectors due to numerical preicision
            norm = np.float64(1)
        else:
            # needed to be able to divide the tuple by 'norm'
            norm = np.float64(sum(outcome.values()))

        # sorting dictionary. adding a callback function to sort by values instead of keys
        # setting reverse = True to be able to obtain the states with highest counts
        outcome_list = sorted(outcome.items(), key=lambda item: item[1], reverse=True)
        states, counts = zip(*outcome_list)

        # normalizing to obtain probabilities
        probs = counts / norm

        # total number of states / number of states with != 0 counts for shot simulators
        total = len(states)

        # number of states that fit without distortion in figure
        upper_bound = 40
        # default fontsize
        font = "medium"

        if n_states_to_keep:
            if n_states_to_keep > total:
                raise ValueError(
                    "n_states_to_keep must be smaller or equal than the total number"
                    f"of states in measurement outcome: {total}"
                )
            else:
                if n_states_to_keep > upper_bound:
                    print("number of states_to_keep exceeds the recommended value")
                    font = "small"

        # if states_to_keep is not given
        else:
            if total > upper_bound:
                n_states_to_keep = upper_bound
            else:
                n_states_to_keep = total

        # formatting labels
        labels = [
            r"$\left|{}\right>$".format(state) for state in states[:n_states_to_keep]
        ]
        labels.append("rest")

        # represent the bar with the addition of all the remaining probabilities
        rest = sum(probs[n_states_to_keep:])

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        colors = [color for _ in range(n_states_to_keep)] + ["xkcd:magenta"]

        ax.bar(labels, np.append(probs[:n_states_to_keep], rest), color=colors)
        ax.set_xlabel("Eigen-State")
        ax.set_ylabel("Probability")
        ax.set_title(label)
        ax.tick_params(axis="x", labelrotation=75, labelsize=font)
        ax.grid(True, axis="y", linestyle="--")

        print("states kept:", n_states_to_keep)
        return

    def plot_n_shots(
        self,
        figsize=(10, 8),
        param_to_plot=None,
        label=None,
        linestyle="--",
        color=None,
        ax=None,
        xlabel="Iterations",
        ylabel="Number of shots",
        title="Evolution of number of shots for gradient estimation",
    ):
        """
        Helper function to plot the evolution of the number of shots used for each evaluation of
        the cost function when computing the gradient.
        It only works for shot adaptive optimizers: cans and icans.
        If cans was used, the number of shots will be the same for each parameter at each iteration.
        If icans was used, the number of shots could be different for each parameter at each iteration.

        Parameters
        ----------
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        param_to_plot: `list[int]` or `int`
            The parameteres to plot. If None, all parameters will be plotted. Defaults to None.
            If a int is given, only the parameter with that index will be plotted.
            If a list of ints is given, the parameters with those indexes will be plotted.
        label: `list[str]` or `str`
            The label for each parameter. Defaults to Parameter {i}.
            If only one parameter is plot the label can be a string, otherwise it must be a list of strings.
        linestyle: `list[str]` or `str`
            The linestyle for each parameter. Defaults to '--' for all parameters.
            If it is a string all parameters will use it, if it a list of strings the linestyle of
            each parameter will depend on one string of the list.
        color: `list[str]` or `str`
            The color for each parameter. Defaults to None for all parameters (matplotlib will choose the colors).
            If only one parameter is plot the color can be a string, otherwise it must be a list of strings.
        ax: 'matplotlib.axes._subplots.AxesSubplot'
            Axis on which to plot the graph. If none is given, a new figure will be created.

        """

        if ax is None:
            ax = plt.subplots(figsize=figsize)[1]

        # creating a list of parameters to plot
        # if param_to_plot is not given, plot all the parameters
        if param_to_plot is None:
            param_to_plot = list(range(len(self.n_shots[0])))
        # if param_to_plot is a single value, convert to list
        elif type(param_to_plot) == int:
            param_to_plot = [param_to_plot]
        # if param_to_plot is not a list, raise error
        if type(param_to_plot) != list:
            raise ValueError(
                "`param_to_plot` must be a list of integers or a single integer"
            )
        else:
            for param in param_to_plot:
                assert param < len(
                    self.n_shots[0]
                ), f"`param_to_plot` must be a list of integers between 0 and {len(self.n_shots[0]) - 1}"

        # if label is not given, create a list of labels for each parameter (only if there is more than 1 parameter)
        if len(self.n_shots[0]) > 1:
            label = (
                [f"Parameter {i}" for i in param_to_plot] if label is None else label
            )
        else:
            label = ["n. shots per parameter"]

        # if only one parameter is plotted, convert label and color to list if they are string
        if len(param_to_plot) == 1:
            if type(label) == str:
                label = [label]
            if type(color) == str:
                color = [color]

        # if param_top_plot is a list and label or color are not lists, raise error
        if (type(label) != list) or (type(color) != list and color is not None):
            raise TypeError("`label` and `color` must be list of str")
        # if label is a list, check that all the elements are strings
        for lab in label:
            assert type(lab) == str, "`label` must be a list of strings"
        # if color is a list, check that all the elements are strings
        if color is not None:
            for c in color:
                assert type(c) == str, "`color` must be a list of strings"

        # if label and color are lists, check if they have the same length as param_to_plot
        if len(label) != len(param_to_plot) or (
            color is not None and len(color) != len(param_to_plot)
        ):
            raise ValueError(
                f"`param_to_plot`, `label` and `color` must have the same length, \
                    `param_to_plot` is a list of {len(param_to_plot)} elements"
            )

        # linestyle must be a string or a list of strings, if it is a string, convert it to a list of strings
        if type(linestyle) != str and type(linestyle) != list:
            raise TypeError("`linestyle` must be str or list")
        elif type(linestyle) == str:
            linestyle = [linestyle for _ in range(len(param_to_plot))]
        elif len(linestyle) != len(param_to_plot):
            raise ValueError(
                f"`linestyle` must have the same length as param_to_plot \
                    (length of `param_to_plot` is {len(param_to_plot)}), or be a string"
            )
        else:
            for ls in linestyle:
                assert type(ls) == str, "`linestyle` must be a list of strings"

        # plot the evolution of the number of shots for each parameter that is in param_to_plot
        transposed_n_shots = np.array(self.n_shots).T
        for i, values in enumerate([transposed_n_shots[j] for j in param_to_plot]):
            if color is None:
                ax.plot(values, label=label[i], linestyle=linestyle[i])
            else:
                ax.plot(values, label=label[i], linestyle=linestyle[i], color=color[i])

        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        ax.legend()
        ax.set_title(title)

    def lowest_cost_bitstrings(self, n_bitstrings: int = 1) -> dict:
        """
        Find the minimium energy from cost_hamilonian given a set of measurement
        outcoms

        Parameters
        ----------

        n_bitstrings : int
            Number of the lowest energies bistrings to get

        Returns
        -------
        best_results : dict
            Returns a list of bitstring with the lowest values of the cost Hamiltonian.

        """

        if isinstance(self.optimized["measurement_outcomes"], dict):
            measurement_outcomes = self.optimized["measurement_outcomes"]
            solution_bitstring = list(measurement_outcomes.keys())
        elif isinstance(self.optimized["measurement_outcomes"], np.ndarray):
            measurement_outcomes = self.get_counts(
                self.optimized["measurement_outcomes"]
            )
            solution_bitstring = list(measurement_outcomes.keys())
        else:
            raise TypeError(
                f"The measurement outcome {type(self.optimized['measurement_outcomes'])} is not valid."
            )
        energies = [
            bitstring_energy(self.cost_hamiltonian, bitstring)
            for bitstring in solution_bitstring
        ]
        args_sorted = np.argsort(energies)
        if n_bitstrings > len(energies):
            n_bitstrings = len(energies)

        total_shots = sum(measurement_outcomes.values())
        best_results = {
            "solutions_bitstrings": [
                solution_bitstring[args_sorted[ii]] for ii in range(n_bitstrings)
            ],
            "bitstrings_energies": [
                energies[args_sorted[ii]] for ii in range(n_bitstrings)
            ],
            "probabilities": [
                measurement_outcomes[solution_bitstring[args_sorted[ii]]] / total_shots
                for ii in range(n_bitstrings)
            ],
        }
        return best_results
    
    def calculate_statistics(self, include_intermediate=False) -> dict:
        """
        A function to calculate statistics of measurement outcomes associated with a QAOA workflow

        Parameters
        ----------
        include_intermediate: `bool`
            Whether it is necessary to calculate statistics for intermediate results. Defaults to False.
        """
        if len(self.intermediate["measurement_outcomes"]) == 0 and include_intermediate == True:
            raise ValueError(
                "The underlying QAOA object does not seem to have any intermediate measurement result. Please, consider saving "
                "intermediate measurements during optimization by setting `optimization_progress=True` in your workflow."
            )

        if isinstance(self.optimized["measurement_outcomes"], dict):
            optimized_measurement_outcomes = self.optimized["measurement_outcomes"]
        elif isinstance(self.optimized["measurement_outcomes"], np.ndarray):
            optimized_measurement_outcomes = self.get_counts(
                self.optimized["measurement_outcomes"]
            )
        else:
            raise TypeError(
                f"The measurement outcome {type(self.optimized['measurement_outcomes'])} is not valid."
            )

        if isinstance(self.intermediate["measurement_outcomes"], list):
            if len(self.intermediate["measurement_outcomes"]) > 0:
                if isinstance(self.intermediate["measurement_outcomes"][0], dict):
                    intermediate_measurement_outcomes = self.intermediate["measurement_outcomes"]
                elif isinstance(self.intermediate["measurement_outcomes"][0], np.ndarray):
                    intermediate_measurement_outcomes = [self.get_counts(i) for i in self.intermediate["measurement_outcomes"]]
                else:
                    raise TypeError(
                        f"The measurement outcome {type(self.intermediate['measurement_outcomes'][0])} is not valid."
                    )
        else:
            raise TypeError(
                f"The measurement outcome {type(self.intermediate['measurement_outcomes'])} is not valid."
            )

        def sorted_mean_std_deviation(counts: dict):
            values = list(counts.values())
            return {
                'sorted': dict(sorted(counts.items(), key=lambda x: x[1], reverse=True)),
                'mean': np.mean(values),
                'std_deviation': np.std(values)
            }

        return {
            'intermediate': [sorted_mean_std_deviation(i) for i in intermediate_measurement_outcomes] if include_intermediate else [],
            'optimized': sorted_mean_std_deviation(optimized_measurement_outcomes)
        }



from typing import Callable

from braket.circuits import gates, Circuit

from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
import openqaoa.qaoa_components.ansatz_constructor.gates as gates_core


class BraketGateApplicator(gates_core.GateApplicator):
    BRAKET_OQ_GATE_MAPPER = {
        gates_core.X.__name__: gates.X.x,
        gates_core.RZ.__name__: gates.Rz.rz,
        gates_core.RX.__name__: gates.Rx.rx,
        gates_core.RY.__name__: gates.Ry.ry,
        gates_core.CX.__name__: gates.CNot.cnot,
        gates_core.CZ.__name__: gates.CZ.cz,
        gates_core.RXX.__name__: gates.XX.xx,
        gates_core.RZZ.__name__: gates.ZZ.zz,
        gates_core.RYY.__name__: gates.YY.yy,
        gates_core.CPHASE.__name__: gates.CPhaseShift.cphaseshift,
        gates_core.RiSWAP.__name__: gates.XY.xy,
    }

    library = "braket"

    def create_quantum_circuit(self, n_qubits) -> Circuit:
        """
        Function which creates and empty circuit for the braket backend.
        Needed for SPAM twirling but more general than that.
        """
        return Circuit()

    def gate_selector(self, gate: gates_core.Gate) -> Callable:
        selected_braket_gate = BraketGateApplicator.BRAKET_OQ_GATE_MAPPER[gate.__name__]
        return selected_braket_gate

    @staticmethod
    def apply_1q_rotation_gate(
        braket_gate, qubit_1: int, rotation_object: RotationAngle, circuit: Circuit
    ) -> Circuit:
        circuit += braket_gate(qubit_1, rotation_object.rotation_angle)
        return circuit

    @staticmethod
    def apply_2q_rotation_gate(
        braket_gate,
        qubit_1: int,
        qubit_2: int,
        rotation_object: RotationAngle,
        circuit: Circuit,
    ) -> Circuit:
        circuit += braket_gate(qubit_1, qubit_2, rotation_object.rotation_angle)
        return circuit

    @staticmethod
    def apply_1q_fixed_gate(braket_gate, qubit_1: int, circuit: Circuit) -> Circuit:
        circuit += braket_gate(qubit_1)
        return circuit

    @staticmethod
    def apply_2q_fixed_gate(
        braket_gate, qubit_1: int, qubit_2: int, circuit: Circuit
    ) -> Circuit:
        circuit += braket_gate(qubit_1, qubit_2)
        return circuit

    def apply_gate(self, gate: gates_core.Gate, *args) -> Circuit:
        selected_braket_gate = self.gate_selector(gate)
        if gate.n_qubits == 1:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,rotation_object,circuit)
                return self.apply_1q_rotation_gate(selected_braket_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,circuit)
                return self.apply_1q_fixed_gate(selected_braket_gate, *args)
        elif gate.n_qubits == 2:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,qubit_2,rotation_object,circuit)
                return self.apply_2q_rotation_gate(selected_braket_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,qubit_2,circuit)
                return self.apply_2q_fixed_gate(selected_braket_gate, *args)
        else:
            raise ValueError("Only 1 and 2-qubit gates are supported.")



from enum import Enum
from typing import Union


class GateMapType(Enum):
    MIXER = "MIXER"
    COST = "COST"
    FIXED = "FIXED"

    @classmethod
    def supported_types(cls):
        return list(map(lambda c: c.name, cls))


class GateMapLabel:
    """
    This object helps keeps track of labels associated with
    gates for their identification in the circuit.
    """

    def __init__(
        self,
        n_qubits: int = None,
        layer_number: int = None,
        application_sequence: int = None,
        gatemap_type: GateMapType = None,
    ):
        """
        Parameters
        ----------
        layer_number: `int`
            The label for the algorthmic layer
        application_sequence: `int`
            The label for the sequence of application
            for the gate
        gate_type: `GateMapType`
            Gate type for distinguishing gate between different QAOA blocks,
            and between parameterized or non-parameterized
        n_qubits: `int`
            Number of qubits in the gate
        """
        if (
            isinstance(layer_number, (int, type(None)))
            and isinstance(application_sequence, (int, type(None)))
            and isinstance(gatemap_type, (GateMapType, type(None)))
        ):
            self.layer = layer_number
            self.sequence = application_sequence
            self.type = gatemap_type
            self.n_qubits = n_qubits
        else:
            raise ValueError("Some or all of input types are incorrect")

    def __repr__(self):
        """
        String representation of the Gatemap label
        """
        representation = f"{self.n_qubits}Q_" if self.n_qubits is not None else ""
        representation += f"{self.type.value}" if self.type.value is not None else ""
        representation += f"_seq{self.sequence}" if self.sequence is not None else ""
        representation += f"_layer{self.layer}" if self.layer is not None else ""

        return representation

    def update_gatelabel(
        self,
        new_layer_number: int = None,
        new_application_sequence: int = None,
        new_gatemap_type: GateMapType = None,
    ) -> None:
        """
        Change the properties of the gatemap label to update
        the gate identity
        """
        if (
            new_layer_number is None
            and new_gatemap_type is None
            and new_application_sequence is None
        ):
            raise ValueError(
                "Pass atleast one updated attribute to update the gatemap label"
            )
        else:
            if isinstance(new_layer_number, int):
                self.layer = new_layer_number
            if isinstance(new_application_sequence, int):
                self.sequence = new_application_sequence
            if isinstance(new_gatemap_type, GateMapType):
                self.type = new_gatemap_type



"""
Tests to run the notebooks. They work by converting the notebooks
to a python script via nbconvert and then running the resulting .py file.
"""
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import pytest

import sys
import os

myPath = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, myPath + "/../")


def notebook_test_function(name):
    with open(name, encoding="utf-8") as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=600, kernel_name="env")

    ep.preprocess(nb)


# @pytest.mark.notebook
def test_01_workflows_example():
    notebook_test_function("./examples/01_workflows_example.ipynb")


# @pytest.mark.notebook
def test_02_simulators_comparison():
    notebook_test_function("./examples/02_simulators_comparison.ipynb")


@pytest.mark.qpu
def test_03_qaoa_on_qpus():
    notebook_test_function("./examples/03_qaoa_on_qpus.ipynb")


# @pytest.mark.notebook
def test_04_qaoa_variational_parameters():
    notebook_test_function("./examples/04_qaoa_variational_parameters.ipynb")


# @pytest.mark.notebook
def test_05_advanced_parameterization():
    notebook_test_function("./examples/05_advanced_parameterization.ipynb")


# @pytest.mark.notebook
def test_06_fast_qaoa_simulator():
    notebook_test_function("./examples/06_fast_qaoa_simulator.ipynb")


# @pytest.mark.notebook
def test_07_cost_landscapes_w_manual_mode():
    notebook_test_function("./examples/07_cost_landscapes_w_manual_mode.ipynb")


# @pytest.mark.notebook
def test_08_results_example():
    notebook_test_function("./examples/08_results_example.ipynb")


def test_09_RQAOA_example():
    notebook_test_function("./examples/09_RQAOA_example.ipynb")


@pytest.mark.qpu
def test_10_workflows_on_Amazon_braket():
    notebook_test_function("./examples/10_workflows_on_Amazon_braket.ipynb")


def test_11_Mixer_example():
    notebook_test_function("./examples/11_Mixer_example.ipynb")


def test_13_optimizers():
    notebook_test_function("./examples/13_optimizers.ipynb")


@pytest.mark.qpu
def test_14_benchmark():
    notebook_test_function("./examples/14_qaoa_benchmark.ipynb")


def test_X_dumping_data():
    notebook_test_function("./examples/X_dumping_data.ipynb")


# Community Tutorials
# @pytest.mark.notebook
def test_tutorial_quantum_approximate_optimization_algorithm():
    notebook_test_function(
        "./examples/community_tutorials/01_tutorial_quantum_approximate_optimization_algorithm.ipynb"
    )


# @pytest.mark.notebook
def test_docplex_example():
    notebook_test_function("./examples/community_tutorials/02_docplex_example.ipynb")


# @pytest.mark.notebook
def test_portfolio_optimization():
    notebook_test_function(
        "./examples/community_tutorials/03_portfolio_optimization.ipynb"
    )


# @pytest.mark.notebook
def test_binpacking():
    notebook_test_function("./examples/community_tutorials/04_binpacking.ipynb")



import unittest

from openqaoa import QAOA
from openqaoa.problems import QUBO
from openqaoa.backends import create_device
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS


class TestSimulators(unittest.TestCase):
    def test_job_ids(self):
        """
        Test if correct job ids are generated and returned for all simulators
        """

        # define problem
        problem = QUBO.random_instance(3)

        # loop over all simulators
        for n in SUPPORTED_LOCAL_SIMULATORS:
            # initialize
            q = QAOA()

            # device
            device = create_device(location="local", name=n)
            q.set_device(device)

            # classical optimizer only 3 iterations
            q.set_classical_optimizer(maxiter=3)

            # compile
            q.compile(problem)

            # run
            q.optimize()

            # check if we have job ids
            opt_id = q.result.optimized["job_id"]
            assert len(opt_id) == 36 and isinstance(
                opt_id, str
            ), f"simulator {n}: job id is not a string of length 36, but {opt_id}"

            inter_id = q.result.intermediate["job_id"]
            for id in inter_id:
                assert len(id) == 36 and isinstance(
                    id, str
                ), f"simulator {n}: on intermediate job id is not a string of length 36, but {id}"


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from typing import Union, Optional, List
import numpy as np

from .plugin_finder import PLUGIN_DICT
from .qaoa_vectorized import QAOAvectorizedBackendSimulator
from .qaoa_analytical_sim import QAOABackendAnalyticalSimulator
from .devices_core import DeviceBase, DeviceLocal
from .basebackend import QuantumCircuitBase, QAOABaseBackend
from ..qaoa_components import QAOADescriptor


def _create_mappers(input_plugin_dict: dict) -> dict:
    DEVICE_NAME_TO_OBJECT_MAPPER = dict()
    DEVICE_ACCESS_OBJECT_MAPPER = dict()

    DEVICE_NAME_TO_OBJECT_MAPPER["vectorized"] = QAOAvectorizedBackendSimulator
    DEVICE_NAME_TO_OBJECT_MAPPER[
        "analytical_simulator"
    ] = QAOABackendAnalyticalSimulator

    for each_entry_key, each_entry_value in input_plugin_dict.items():
        if hasattr(each_entry_value, "device_access"):
            DEVICE_ACCESS_OBJECT_MAPPER.update(each_entry_value.device_access)
        if hasattr(each_entry_value, "device_name_to_obj"):
            DEVICE_NAME_TO_OBJECT_MAPPER.update(each_entry_value.device_name_to_obj)

    return DEVICE_NAME_TO_OBJECT_MAPPER, DEVICE_ACCESS_OBJECT_MAPPER


DEVICE_NAME_TO_OBJECT_MAPPER, DEVICE_ACCESS_OBJECT_MAPPER = _create_mappers(PLUGIN_DICT)


def _backend_arg_mapper(
    backend_obj: QAOABaseBackend,
    n_shots: Optional[int] = None,
    seed_simulator: Optional[int] = None,
    qiskit_simulation_method: Optional[str] = None,
    qiskit_optimization_level: Optional[int] = None,
    noise_model=None,
    active_reset: Optional[bool] = None,
    rewiring=None,
    disable_qubit_rewiring: Optional[bool] = None,
    initial_qubit_mapping=None,
):
    BACKEND_ARGS_MAPPER = {
        QAOABackendAnalyticalSimulator: {},
        QAOAvectorizedBackendSimulator: {},
    }

    local_vars = locals()

    for each_plugin_entrypoint in PLUGIN_DICT.values():
        if hasattr(each_plugin_entrypoint, "backend_args"):
            for each_key, each_value in each_plugin_entrypoint.backend_args.items():
                # Convert list of accepted parameters into a dictionary with
                # the name of the variable as a key and the local value of the
                # variable
                var_values = [local_vars[each_name] for each_name in each_value]
                input_dict = {each_key: dict(zip(each_value, var_values))}
                BACKEND_ARGS_MAPPER.update(input_dict)

    final_backend_kwargs = {
        key: value
        for key, value in BACKEND_ARGS_MAPPER[backend_obj].items()
        if value is not None
    }
    return final_backend_kwargs


def device_to_backend_mapper(device: DeviceBase) -> QAOABaseBackend:
    """
    Return the correct `QAOABaseBackend` object corresponding to the
    requested device
    """
    if isinstance(device, DeviceLocal):
        try:
            backend_class = DEVICE_NAME_TO_OBJECT_MAPPER[device.device_name]
        except KeyError:
            raise ValueError(
                f"The device {device} is not supported."
                f"Please choose from {DEVICE_NAME_TO_OBJECT_MAPPER.keys()}"
            )
        except:
            raise Exception(f"The device name {device} raised an unknown error")

    else:
        try:
            backend_class = DEVICE_ACCESS_OBJECT_MAPPER[type(device)]
        except KeyError:
            raise ValueError(
                f"The device {device} is not supported."
                f"Please choose from {DEVICE_ACCESS_OBJECT_MAPPER.keys()}"
            )
        except:
            raise Exception(f"The specified {device} raised an unknown error")

    return backend_class


def get_qaoa_backend(
    qaoa_descriptor: QAOADescriptor,
    device: DeviceBase,
    prepend_state: Optional[
        Union[QuantumCircuitBase, List[complex], np.ndarray]
    ] = None,
    append_state: Optional[Union[QuantumCircuitBase, np.ndarray]] = None,
    init_hadamard: bool = True,
    cvar_alpha: float = 1,
    **kwargs,
):
    """
    A wrapper function to return a QAOA backend object.

    Parameters
    ----------
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction along with depth of the circuit.
    device: `DeviceBase`
        The device to be used: Specified as an object of the class `DeviceBase`.
    prepend_state: `Union[QuantumCircuitBase,np.ndarray(complex)]`
        The state prepended to the circuit.
    append_state: `Union[QuantumCircuitBase,np.ndarray(complex)]`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.
    cvar_alpha: `float`
        The value of the CVaR parameter.
    kwargs:
    Additional keyword arguments for the backend.
        initial_qubit_mapping: `list`
            A list of physical qubits to be used for the QAOA circuit.
        n_shots: `int`
            The number of shots to be used for the shot-based computation.

    Returns
    -------
    `QAOABaseBackend`
        The corresponding backend object.
    """

    backend_class = device_to_backend_mapper(device)
    backend_kwargs = _backend_arg_mapper(backend_class, **kwargs)

    try:
        if isinstance(device, DeviceLocal):
            backend_obj = backend_class(
                qaoa_descriptor=qaoa_descriptor,
                prepend_state=prepend_state,
                append_state=append_state,
                init_hadamard=init_hadamard,
                cvar_alpha=cvar_alpha,
                **backend_kwargs,
            )

        else:
            backend_obj = backend_class(
                qaoa_descriptor=qaoa_descriptor,
                device=device,
                prepend_state=prepend_state,
                append_state=append_state,
                init_hadamard=init_hadamard,
                cvar_alpha=cvar_alpha,
                **backend_kwargs,
            )

    except Exception as e:
        raise ValueError(f"The backend returned an error: {e}")
    return backend_obj



import unittest
import importlib
import os

from setuptools import find_namespace_packages


class TestImports(unittest.TestCase):

    """
    Test module imports within OpenQAOA
    """

    def test_all_module_import(self):
        """
        Test all the main module imports for OQ
        """
        
        folder_names = [each_file for each_file in os.listdir('src') if 'openqaoa-' in each_file]
        
        packages_import = []
        for each_file_name in folder_names:
            packages_import.extend(find_namespace_packages(where="./src/"+each_file_name, exclude=['dist', 'build', 'build.*', 'tests', 'tests.*']))
        print(packages_import)
        packages_import = sorted(packages_import)
        for each_package in packages_import:
            try:
                importlib.import_module(each_package)
            except Exception as e:
                raise Exception(e)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""SPSA optimizer"""

from openqaoa.optimizers.pennylane import (
    numpy as np,
)  # changed from the original PennyLane code


class SPSAOptimizer:
    r"""The Simultaneous Perturbation Stochastic Approximation method (SPSA)
    is a stochastic approximation algorithm for optimizing cost functions whose evaluation may involve noise.

    While other gradient-based optimization methods usually attempt to compute
    the gradient analytically, SPSA involves approximating gradients at the cost of
    evaluating the cost function twice in each iteration step. This cost may result in
    a significant decrease in the overall cost of function evaluations for the entire optimization.
    It is based on an approximation of the unknown gradient :math:`\hat{g}(\hat{\theta}_{k})`
    through a simultaneous perturbation of the input parameters:

    .. math::
        \hat{g}_k(\hat{\theta}_k) = \frac{y(\hat{\theta}_k+c_k\Delta_k)-
        y(\hat{\theta}_k-c_k\Delta_k)}{2c_k} \begin{bmatrix}
           \Delta_{k1}^{-1} \\
           \Delta_{k2}^{-1} \\
           \vdots \\
           \Delta_{kp}^{-1}
         \end{bmatrix}\text{,}

    where

    * :math:`k` is the current iteration step,
    * :math:`\hat{\theta}_k` are the input parameters at iteration step :math:`k`,
    * :math:`y` is the objective function,
    * :math:`c_k=\frac{c}{k^\gamma}` is the gain sequence corresponding to evaluation step size
      and it can be controlled with

      * scaling parameter :math:`c` and
      * scaling exponent :math:`\gamma`

    * :math:`\Delta_{ki}^{-1} \left(1 \leq i \leq p \right)` are the inverted elements of
      random pertubation vector :math:`\Delta_k`.

    :math:`\hat{\theta}_k` is updated to a new set of parameters with

    .. math::
        \hat{\theta}_{k+1} = \hat{\theta}_{k} - a_k\hat{g}_k(\hat{\theta}_k)\text{,}

    where the gain sequences :math:`a_k=\frac{a}{(A+k)^\alpha}` controls parameter update step size.

    The gain sequence :math:`a_k` can be controlled with

    * scaling parameter :math:`a`,
    * scaling exponent :math:`\alpha` and
    * stability constant :math:`A`

    For more details, see `Spall (1998a)
    <https://www.jhuapl.edu/SPSA/PDF-SPSA/Spall_An_Overview.PDF>`_.

    .. note::

        * One SPSA iteration step of a cost function that involves computing the expectation value of
          a Hamiltonian with ``M`` terms requires :math:`2*M` quantum device executions.
        * The forward-pass value of the cost function is not computed when stepping the optimizer.
          Therefore, in case of using ``step_and_cost`` method instead of ``step``, the number
          of executions will include the cost function evaluations.


    **Examples:**

    For VQE/VQE-like problems, the objective function can be the following:

    >>> coeffs = [0.2, -0.543, 0.4514]
    >>> obs = [qml.PauliX(0) @ qml.PauliZ(1), qml.PauliZ(0) @ qml.Hadamard(2),
    ...             qml.PauliX(3) @ qml.PauliZ(1)]
    >>> H = qml.Hamiltonian(coeffs, obs)
    >>> num_qubits = 4
    >>> dev = qml.device("default.qubit", wires=num_qubits)
    >>> @qml.qnode(dev)
    ... def cost(params, num_qubits=1):
    ...     qml.BasisState(np.array([1, 1, 0, 0]), wires=range(num_qubits))
    ...     for i in range(num_qubits):
    ...         qml.Rot(*params[i], wires=0)
    ...         qml.CNOT(wires=[2, 3])
    ...         qml.CNOT(wires=[2, 0])
    ...         qml.CNOT(wires=[3, 1])
    ...     return qml.expval(H)
    ...
    >>> params = np.random.normal(0, np.pi, (num_qubits, 3), requires_grad=True)

    Once constructed, the cost function can be passed directly to the
    ``step`` or ``step_and_cost`` function of the optimizer:

    >>> max_iterations = 100
    >>> opt = qml.SPSAOptimizer(maxiter=max_iterations)
    >>> for _ in range(max_iterations):
    ...     params, energy = opt.step_and_cost(cost, params, num_qubits=num_qubits)
    >>> print(energy)
    -0.4294539602541956

    The algorithm provided by SPSA does not rely on built-in automatic differentiation capabilities of the interface being used
    and therefore the optimizer can be used in more complex hybrid classical-quantum workflow with any of the interfaces:

    >>> n_qubits = 1
    >>> max_iterations = 20
    >>> dev = qml.device("default.qubit", wires=n_qubits)
    >>> @qml.qnode(dev, interface="tf")
    ... def layer_fn_spsa(inputs, weights):
    ...     qml.AngleEmbedding(inputs, wires=range(n_qubits))
    ...     qml.BasicEntanglerLayers(weights, wires=range(n_qubits))
    ...     return qml.expval(qml.PauliZ(wires=0))
    ...
    >>> opt = qml.SPSAOptimizer(maxiter=max_iterations)
    ... def fn(params, tensor_in, tensor_out):
    ...     with tf.init_scope():
    ...             for _ in range(max_iterations):
    ...                     # Some classical steps before the quantum computation
    ...                     params_a, layer_res = opt.step_and_cost(layer_fn_spsa,
    ...                                     np.tensor(tensor_in, requires_grad=False),
    ...                                     np.tensor(params))
    ...                     params = params_a[1]
    ...                     tensor_out = layer_res
    ...                     # Some classical steps after the quantum computation
    ...     return layer_res
    ...
    >>> tensor_in = tf.Variable([0.27507603])
    >>> tensor_out = tf.Variable([0])
    >>> params = tf.Variable([[3.97507603],
    ...     [3.12950603],
    ...     [1.00854038],
    ...     [1.25907603]])
    >>> loss = fn(params, tensor_in, tensor_out)
    >>> print(loss)
    tf.Tensor(-0.9995854230771829, shape=(), dtype=float64)



    Keyword Args:
        maxiter (int): the maximum number of iterations expected to be performed.
            Used to determine :math:`A`, if :math:`A` is not supplied, otherwise ignored.
        alpha (float): A hyperparameter to calculate :math:`a_k=\frac{a}{(A+k+1)^\alpha}`
            for each iteration. Its asymptotically optimal value is 1.0.
        gamma (float): An hyperparameter to calculate :math:`c_k=\frac{c}{(k+1)^\gamma}`
            for each iteration. Its asymptotically optimal value is 1/6.
        c (float): A hyperparameter related to the expected noise. It should be
            approximately the standard deviation of the expected noise of the cost function.
        A (float): The stability constant; if not provided, set to be 10% of the maximum number
            of expected iterations.
        a (float): A hyperparameter expected to be small in noisy situations,
            its value could be picked using `A`, :math:`\alpha` and :math:`\hat{g_0} (\hat{\theta_0})`.
            For more details, see `Spall (1998b)
            <https://www.jhuapl.edu/spsa/PDF-SPSA/Spall_Implementation_of_the_Simultaneous.PDF>`_.
    """

    # pylint: disable-msg=too-many-arguments
    def __init__(self, maxiter=None, alpha=0.602, gamma=0.101, c=0.2, A=None, a=None):
        self.a = a
        self.A = A
        if not maxiter and not A:
            raise TypeError("One of the parameters maxiter or A must be provided.")
        if not A:
            self.A = maxiter * 0.1
        if not a:
            self.a = 0.05 * (self.A + 1) ** alpha
        self.c = c
        self.alpha = alpha
        self.gamma = gamma
        self.k = 1
        self.ak = self.a / (self.A + 1) ** self.alpha

    def step_and_cost(self, objective_fn, *args, **kwargs):
        r"""Update the parameter array :math:`\hat{\theta}_k` with one step of the
        optimizer and return the step and the corresponding objective function. The number
        of steps stored by the ``k`` attribute of the optimizer is counted internally when calling ``step_and_cost`` and ``cost``.

        Args:
            objective_fn (function): the objective function for optimization
            *args : variable length argument array for objective function
            **kwargs : variable length of keyword arguments for the objective function

        Returns:
            tuple[list [array], float]: the new variable values :math:`\hat{\theta}_{k+1}` and the
            objective function output prior to the step.
        """
        g = self.compute_grad(objective_fn, args, kwargs)
        new_args = self.apply_grad(g, args)

        self.k += 1

        forward = objective_fn(*args, **kwargs)

        # unwrap from list if one argument, cleaner return
        if len(new_args) == 1:
            return new_args[0], forward
        return new_args, forward

    def step(self, objective_fn, *args, **kwargs):
        r"""Update trainable arguments with one step of the optimizer. The number
        of steps is being counted through calls to ``step_and_cost`` and ``cost``.

        Args:
            objective_fn (function): the objective function for optimization
            *args : variable length argument array for objective function
            **kwargs : variable length of keyword arguments for the objective function

        Returns:
            list [array]: the new variable values :math:`\hat{\theta}_{k+1}`.
        """
        g = self.compute_grad(objective_fn, args, kwargs)
        new_args = self.apply_grad(g, args)

        self.k += 1

        # unwrap from list if one argument, cleaner return
        if len(new_args) == 1:
            return new_args[0]

        return new_args

    def compute_grad(self, objective_fn, args, kwargs):
        r"""Approximate the gradient of the objective function at the
        given point.

        Args:
            objective_fn (function): The objective function for optimization
            args (tuple): tuple of NumPy array containing the current parameters
                for objective function
            kwargs (dict): keyword arguments for the objective function

        Returns:
            tuple (array): NumPy array containing the gradient
                :math:`\hat{g}_k(\hat{\theta}_k)`
        """
        ck = self.c / self.k**self.gamma

        delta = []
        thetaplus = list(args)
        thetaminus = list(args)

        for index, arg in enumerate(args):
            if getattr(arg, "requires_grad", False):
                # Use the symmetric Bernoulli distribution to generate
                # the coordinates of delta. Note that other distributions
                # may also be used (they need to satisfy certain conditions).
                # Refer to the paper linked in the class docstring for more info.
                di = np.random.choice([-1, 1], size=arg.shape)
                multiplier = ck * di
                thetaplus[index] = arg + multiplier
                thetaminus[index] = arg - multiplier
                delta.append(di)
        yplus = objective_fn(*thetaplus, **kwargs)
        yminus = objective_fn(*thetaminus, **kwargs)
        try:
            if np.prod(objective_fn.func(*args).shape(objective_fn.device)) > 1:
                raise ValueError(
                    "The objective function must be a scalar function for the gradient "
                    "to be computed."
                )
        except AttributeError:
            if yplus.size > 1:
                raise ValueError(  # pylint: disable=raise-missing-from
                    "The objective function must be a scalar function for the gradient "
                    "to be computed."
                )
        grad = [(yplus - yminus) / (2 * ck * di) for di in delta]

        return tuple(grad)

    def apply_grad(self, grad, args):
        r"""Update the variables to take a single optimization step.

        Args:
            grad (tuple [array]): the gradient approximation of the objective
                function at point :math:`\hat{\theta}_{k}`
            args (tuple): the current value of the variables :math:`\hat{\theta}_{k}`

        Returns:
            list [array]: the new values :math:`\hat{\theta}_{k+1}`"""
        self.ak = self.a / (self.A + self.k) ** self.alpha
        args_new = list(args)
        trained_index = 0
        for index, arg in enumerate(args):
            if getattr(arg, "requires_grad", False):
                args_new[index] = arg - self.ak * grad[trained_index]
                trained_index += 1

        return args_new



import networkx as nx

from ..utilities import check_kwargs
from .problem import Problem
from .qubo import QUBO


class ShortestPath(Problem):
    """
    Creates an instance of the Shortest Path problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.
    source: int
        The index of the source node.
    dest: int
        The index of the destination node.

    Returns
    -------
        An instance of the Shortest Path problem.
    """

    __name__ = "shortest_path"

    def __init__(self, G, source, dest):
        # Relabel nodes to integers starting from 0
        mapping = dict(zip(G, range(G.number_of_nodes())))
        self.G = nx.relabel_nodes(G, mapping)

        self.source = source
        self.dest = dest

        assert source in list(G.nodes), f"Source node not within nodes of input graph"
        assert dest in list(
            G.nodes
        ), f"Destination node not within nodes of input graph"
        assert source != dest, "Source and destination nodes cannot be the same"

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Shortest problem, whose graph is
        random following the Erdos-Renyi model. By default the node and edge
        weights are set to 1.0 and the default constraint is taken to be as large.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
        A random instance of the Shortest Path problem.
        """

        n_nodes, edge_probability, seed, source, dest = check_kwargs(
            ["n_nodes", "edge_probability", "seed", "source", "dest"],
            [None, None, 1234, 0, 1],
            **kwargs,
        )
        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )

        DEFAULT_WEIGHTS = 1.0

        for u, v in G.edges():
            G.edges[u, v]["weight"] = DEFAULT_WEIGHTS
        for w in G.nodes():
            G.nodes[w]["weight"] = DEFAULT_WEIGHTS

        return ShortestPath(G, source, dest)

    def terms_and_weights(self):
        """
        Creates the terms and weights for the Shortest Path problem

        Returns
        -------
        terms_weights: tuple(list[list],list[float])
            Tuple containing list of terms and list of weights
        """
        s = self.source
        d = self.dest
        n_nodes = self.G.number_of_nodes()
        n_edges = self.G.number_of_edges()

        # # Linear terms due to node weights
        #     # For loop version
        #     node_terms_weights = []
        #     for i in range(n_nodes):
        #         if i not in [s, d]:
        #             shift = int(i>s)+int(i>d)
        #             node_terms_weights.append(([i-shift], self.G.nodes[i]['weight']))
        node_terms_weights = [
            ([i - (int(i > s) + int(i > d))], self.G.nodes[i]["weight"])
            for i in range(n_nodes)
            if i not in [s, d]
        ]

        # Linear terms due to edge weights (shift of n_nodes-2 since we removed 2 nodes)
        #     # For loop version
        #     edge_terms_weights = []
        #     for i, (u,v) in enumerate(self.G.edges()):
        #         edge_terms_weights.append(([i+n_nodes-2], self.G.edges[u,v]['weights']))
        edge_terms_weights = [
            ([i + n_nodes - 2], self.G.edges[u, v]["weight"])
            for i, (u, v) in enumerate(self.G.edges())
        ]

        # Source flow constraint
        #     # For loop version
        #     start_flow_terms_weights = []
        #     for i, x in enumerate(self.G.edges()):
        #         for j, y in enumerate(self.G.edges()):
        #             if s in x and s in y:
        #                 if i == j:
        #                     start_flow_terms_weights.append(([i+n_nodes-2], -1))
        #                 else:
        #                     start_flow_terms_weights.append(([i+n_nodes-2,j+n_nodes-2], 1))
        start_flow_terms_weights = [
            ([i + n_nodes - 2], -1)
            if i == j
            else ([i + n_nodes - 2, j + n_nodes - 2], 1)
            for i, x in enumerate(self.G.edges())
            for j, y in enumerate(self.G.edges())
            if (s in x and s in y)
        ]

        # Destination flow constraint
        #     # For loop version
        #     dest_flow_terms_weights = []
        #     for i, x in enumerate(self.G.edges()):
        #         for j, y in enumerate(self.G.edges()):
        #             if d in x and d in y:
        #                 if i == j:
        #                     dest_flow_terms_weights.append(([i+n_nodes-2], -1))
        #                 else:
        #                     dest_flow_terms_weights.append(([i+n_nodes-2,j+n_nodes-2], 1))
        dest_flow_terms_weights = [
            ([i + n_nodes - 2], -1)
            if i == j
            else ([i + n_nodes - 2, j + n_nodes - 2], 1)
            for i, x in enumerate(self.G.edges())
            for j, y in enumerate(self.G.edges())
            if (d in x and d in y)
        ]

        # Path flow constraint
        path_flow_terms_weights = []
        for i in range(n_nodes):
            if i != d and i != s:
                shift = int(i > s) + int(i > d)
                path_flow_terms_weights.append(([i - shift], 4))
                for j, x in enumerate(self.G.edges()):
                    if i in x:
                        path_flow_terms_weights.append(
                            ([i - shift, j + n_nodes - 2], -4)
                        )
                    for k, y in enumerate(self.G.edges()):
                        if i in x and i in y:
                            if j == k:
                                path_flow_terms_weights.append(([j + n_nodes - 2], 1))
                            else:
                                path_flow_terms_weights.append(
                                    ([j + n_nodes - 2, k + n_nodes - 2], 1)
                                )

        return tuple(
            zip(
                *(
                    node_terms_weights
                    + edge_terms_weights
                    + start_flow_terms_weights
                    + dest_flow_terms_weights
                    + path_flow_terms_weights
                )
            )
        )

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
        The QUBO encoding of this problem.
        """
        n = self.G.number_of_nodes() + self.G.number_of_edges() - 2
        # Extract terms and weights from the problem definition
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        return QUBO(n, ising_terms, ising_weights, self.problem_instance)



from typing import List

from azure.quantum.qiskit import AzureQuantumProvider
from openqaoa.backends.devices_core import DeviceBase


class DeviceAzure(DeviceBase):
    """
    Contains the required information and methods needed to access remote
    Azure QPUs and Simulators.

    Parameters
    ----------
    available_qpus: `list`
        When connection to a provider is established, this attribute contains a list
        of backend names which can be used to access the selected backend by reinitialising
        the Access Object with the name of the available backend as input to the
        device_name parameter.
    """

    def __init__(self, device_name: str, resource_id: str, az_location: str):
        """
        Input parameters required for this can be found in the user's Azure
        Quantum Workspace.

        Parameters
        ----------
        device_name: `str`
            The name of the Azure remote QPU/Simulator to be used
        resource_id: `str`
            The resource_id of the Workplace
        az_location: `str`
            The location of the Azure Workplace. e.g. "westus"
        """

        self.resource_id = resource_id
        self.location = az_location
        self.device_name = device_name
        self.device_location = "azure"

        self.provider_connected = None
        self.qpu_connected = None

    def check_connection(self):
        """ """

        self.provider_connected = self._check_provider_connection()

        if self.provider_connected == False:
            return self.provider_connected

        self.available_qpus = [backend.name() for backend in self.provider.backends()]

        if self.device_name == "":
            return self.provider_connected

        self.qpu_connected = self._check_backend_connection()

        if self.provider_connected and self.qpu_connected:
            return True
        else:
            return False

    def _check_backend_connection(self) -> bool:
        """Private method for checking connection with backend(s)."""

        if self.device_name in self.available_qpus:
            self.backend_device = self.provider.get_backend(self.device_name)
            self.n_qubits = self.backend_device.configuration().n_qubits
            return True
        else:
            print(f"Please choose from {self.available_qpus} for this provider")
            return False

    def _check_provider_connection(self) -> bool:
        """
        Private method for checking connection with provider.
        """

        try:
            self.provider = AzureQuantumProvider(
                resource_id=self.resource_id, location=self.location
            )

            return True

        except ValueError as e:
            print(
                "Either the resource id or location specified was invalid: {}".format(e)
            )
            return False

        except Exception as e:
            print(
                "An Exception has occured when trying to connect with the \
            provider: {}".format(
                    e
                )
            )
            return False

    def connectivity(self) -> List[List[int]]:
        return self.backend_device.configuration().coupling_map



import unittest
import itertools
import pytest

from openqaoa_qiskit.backends import DeviceQiskit


class TestingDeviceQiskit(unittest.TestCase):

    """These tests check the Object used to access IBMQ and their available
    QPUs can be established.
    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with IBMQ.

    Please ensure that the provided api token, hub, group and project in the
    crendentials.json are correct.
    All of these can be found in your IBMQ Account Page.
    """

    @pytest.mark.api
    def setUp(self):

        self.HUB = "ibm-q"
        self.GROUP = "open"
        self.PROJECT = "main"
        self.INSTANCE = "ibm-q/open/main"

    @pytest.mark.api
    def test_changing_provider(self):

        """
        This test checks that the specified hub,group and project in the
        initialisation of DeviceQiskit changes the provider to the appropriate
        destination.
        """

        device_obj = DeviceQiskit(device_name="ibmq_manila")
        device_obj.check_connection()

        provider_instances = device_obj.provider.instances()

        if len(provider_instances) >= 2:

            for each_item in provider_instances[:2]:

                [hub, group, project] = each_item.split("/")
                device_obj2 = DeviceQiskit(
                    device_name="ibmq_manila", hub=hub, group=group, project=project
                )
                device_obj2.check_connection()

                self.assertEqual(device_obj2.provider._account.instance, each_item)

    @pytest.mark.api
    def test_check_connection_provider_no_backend_wrong_hub_group_project(self):

        """
        Hub, group and project must always be specified together.
        If either the hub, group or project is wrongly specified, check_connection should
        return False.
        If not all 3 are specified, check_connection should return False.
        The provider_connected attribute should be updated to False.
        Since the API Token is loaded from save_account, the api token will be
        checked by Qiskit.
        """

        for each_combi in itertools.product(
            ["invalid_hub", None], ["invalid_group", None], ["invalid_project", None]
        ):

            if each_combi != (None, None, None):

                device_obj = DeviceQiskit(
                    device_name="",
                    hub=each_combi[0],
                    group=each_combi[1],
                    project=each_combi[2],
                )

                self.assertEqual(device_obj.check_connection(), False)
                self.assertEqual(device_obj.provider_connected, False)
                self.assertEqual(device_obj.qpu_connected, None)

    @pytest.mark.api
    def test_check_connection_provider_no_backend_provided_credentials(self):

        """
        If no information about the device name, but the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceQiskit(
            device_name="", hub=self.HUB, group=self.GROUP, project=self.PROJECT
        )

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)

    @pytest.mark.api
    def test_check_connection_provider_right_backend_provided_credentials(self):

        """
        If the correct device name is provided and the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceQiskit(
            device_name="", hub=self.HUB, group=self.GROUP, project=self.PROJECT
        )

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceQiskit(
            device_name=valid_qpu_name,
            hub=self.HUB,
            group=self.GROUP,
            project=self.PROJECT,
        )

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)

    @pytest.mark.api
    def test_check_connection_provider_wrong_backend_provided_credentials(self):

        """
        If device name provided is incorrect, and not empty, and the credentials
        used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceQiskit(
            device_name="random_invalid_backend",
            hub=self.HUB,
            group=self.GROUP,
            project=self.PROJECT,
        )

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)


if __name__ == "__main__":
    unittest.main()



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import TSP


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool

class TestTSP(unittest.TestCase):
    """Tests for TSP class"""
    def test_tsp_terms_weights_constant(self):
        """Testing TSP problem creation"""
        city_coordinates = [(4, 1), (4, 4), (3, 3), (1, 3.5)]
        expected_terms = [
            [0, 3],
            [1, 4],
            [2, 5],
            [0, 6],
            [1, 7],
            [8, 2],
            [3, 6],
            [4, 7],
            [8, 5],
            [0, 1],
            [0, 2],
            [1, 2],
            [3, 4],
            [3, 5],
            [4, 5],
            [6, 7],
            [8, 6],
            [8, 7],
            [0, 4],
            [1, 3],
            [3, 7],
            [4, 6],
            [0, 5],
            [2, 3],
            [8, 3],
            [5, 6],
            [1, 5],
            [2, 4],
            [8, 4],
            [5, 7],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]
        expected_weights = [
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            3.905124837953327,
            0.3535533905932738,
            0.3535533905932738,
            0.7603453162872774,
            0.7603453162872774,
            0.3535533905932738,
            0.3535533905932738,
            0.5153882032022076,
            0.5153882032022076,
            0.7603453162872774,
            0.7603453162872774,
            0.5153882032022076,
            0.5153882032022076,
            -10.424148382787205,
            -9.797225258452029,
            -11.038545614372802,
            -10.038047089667756,
            -9.548132863497614,
            -10.361716714885624,
            -10.424148382787205,
            -9.797225258452029,
            -11.038545614372802,
        ]
        expected_constant = 62.51983851122417
        tsp_qubo = TSP(city_coordinates).qubo
        self.assertTrue(terms_list_equality(expected_terms, tsp_qubo.terms))
        self.assertEqual(expected_weights, tsp_qubo.weights)
        self.assertEqual(expected_constant, tsp_qubo.constant)

    def test_tsp_random_instance(self):
        """Testing the random_instance method of the TSP problem class"""
        rng = np.random.default_rng(1234)
        n_cities = 4

        box_size = np.sqrt(n_cities)
        city_coordinates = list(map(tuple, box_size * rng.random(size=(n_cities, 2))))

        tsp_prob = TSP(city_coordinates).qubo

        tsp_prob_random = TSP.random_instance(n_cities=n_cities, seed=1234).qubo

        self.assertTrue(terms_list_equality(tsp_prob_random.terms, tsp_prob.terms))
        self.assertEqual(tsp_prob_random.weights, tsp_prob.weights)
        self.assertEqual(tsp_prob_random.constant, tsp_prob.constant)

    def test_tsp_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """
        # If nothing is given, must return a ValueError
        with self.assertRaises(ValueError) as e:
            TSP()
        self.assertEqual(
            "Input missing: city coordinates, distance matrix or (weighted graph) required",
            str(e.exception),
        )

        # coordinates type-check
        coordinates_list = [(1, 2), {"test": "oh", "test1": "oh"}, np.array([1, 2])]

        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual("The coordinates should be a list", str(e.exception))

        coordinates_list = [[[1, 2], [2, 1]], [np.array([1, 2]), np.array([2, 1])]]
        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual(
                "The coordinates should be contained in a tuple", str(e.exception)
            )

        coordinates_list = [
            [("oh", "num"), ("num", "oh")],
            [(np.array(1), np.array(2)), (np.array(2), np.array(1))],
        ]
        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual(
                "The coordinates must be of type float or int", str(e.exception)
            )

        # coordinates type-check
        distance_matrices = [
            (1, 2),
            np.array([[1, 2], [3, 4]]),
            {"test": "oh", "test1": "oh"},
        ]

        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual("The distance matrix should be a list", str(e.exception))

        # Distance matrix type-check
        distance_matrices = [[(1, 2), (2, 1)], [np.array([1, 2]), np.array([2, 1])]]
        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual(
                "Each row in the distance matrix should be a list", str(e.exception)
            )

        distance_matrices = [
            [["oh", "num"], ["num", "oh"]],
            [[np.array(1), np.array(2)], [np.array(2), np.array(1)]],
        ]
        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual(
                "The distance matrix entries must be of type float or int",
                str(e.exception),
            )

        distance_matrix = [[1, 2.3], [-2, 3]]
        with self.assertRaises(ValueError) as e:
            TSP(distance_matrix=distance_matrix)
        self.assertEqual("Distances should be positive", str(e.exception))

        # Graph type-check
        G = nx.complete_graph(5)
        for u, v in G.edges():
            G[u][v]["weight"] = "a"

        with self.assertRaises(TypeError) as e:
            TSP(G=G)
        self.assertEqual(
            "The edge weights must be of type float or int", str(e.exception)
        )

        for u, v in G.edges():
            G[u][v]["weight"] = -2.0

        with self.assertRaises(ValueError) as e:
            TSP(G=G)
        self.assertEqual("Edge weights should be positive", str(e.exception))


if __name__ == "__main__":
    unittest.main()


import os

from openqaoa_braket.backends import DeviceAWS

from openqaoa.algorithms import QAOA, RQAOA
from openqaoa.algorithms.baseworkflow import Workflow
from openqaoa.backends.qaoa_device import create_device


class AWSJobs(Workflow):
    """
    This class is meant to be used *only* within the framework of AWS managed hybrid jobs.
    """

    def __init__(self, algorithm: str = "qaoa"):
        """
        Initialize the QAOA class.

        Parameters
        ----------
            algorithm: str
                The algorithm to be used. Chose between QAOA and RQAOA
        """
        # the input data directory opt/braket/input/data
        self.input_dir = os.environ["AMZN_BRAKET_INPUT_DIR"]
        # the output directory opt/braket/model to write ob results to
        self.results_dir = os.environ["AMZN_BRAKET_JOB_RESULTS_DIR"]
        # the name of the job
        self.job_name = os.environ["AMZN_BRAKET_JOB_NAME"]
        # the checkpoint directory
        self.checkpoint_dir = os.environ["AMZN_BRAKET_CHECKPOINT_DIR"]
        # the hyperparameter
        self.hyperparameter_file_name = os.environ["AMZN_BRAKET_HP_FILE"]
        # the device ARN (AWS Resource Number)
        self.device_arn = os.environ["AMZN_BRAKET_DEVICE_ARN"]
        # the output S3 bucket, as specified in the CreateJob request’s OutputDataConfig
        self.out_s3_bucket = os.environ["AMZN_BRAKET_OUT_S3_BUCKET"]
        # the entry point as specified in the CreateJob request’s ScriptModeConfig
        self.script_entry_point = os.environ["AMZN_BRAKET_SCRIPT_ENTRY_POINT"]
        # the compression type as specified in the CreateJob request’s ScriptModeConfig
        self.compression_type = os.environ["AMZN_BRAKET_SCRIPT_COMPRESSION_TYPE"]
        # the S3 location of the user’s script as specified in the CreateJob request’s ScriptModeConfig
        self.script_sr_uri = os.environ["AMZN_BRAKET_SCRIPT_S3_URI"]
        # the S3 location where the SDK would store the task results by default for the job
        self.task_results_se_uri = os.environ["AMZN_BRAKET_TASK_RESULTS_S3_URI"]
        # the S3 location where the job results would be stored, as specified in CreateJob request’s OutputDataConfig
        self.results_s3_path = os.environ["AMZN_BRAKET_JOB_RESULTS_S3_PATH"]
        # the string that should be passed to CreateQuantumTask’s jobToken parameter for quantum tasks created in the job container
        # if os.environ["AMZN_BRAKET_JOB_TOKEN"]:
        #     self.job_token = os.environ["AMZN_BRAKET_JOB_TOKEN"]

        self.device = DeviceAWS(
            device_name=self.device_arn,
            folder_name=self.results_dir,
            s3_bucket_name=self.out_s3_bucket,
        )
        self.algorithm = algorithm.lower()
        self.completed = False

    def load_compile_data(self):
        """
        A helper method to load the parameters.

        .. note::
            Note tha this function is executed within the AWS job-docker
        """

        path = os.path.join(self.input_dir, "input_data/")

        if "qaoa" == self.algorithm.lower():
            workflow = QAOA()
        elif "rqaoa" == self.algorithm.lower():
            workflow = RQAOA()
        else:
            raise ValueError(
                f"Specified algorithm {self.algorithm} is not supported. Please choose between [QAOA, RQAOA]"
            )

        self.workflow = workflow.load(file_name="openqaoa_params.json", file_path=path)

        # Extract the original uuid
        self.atomic_uuid = self.workflow.asdict()["header"]["atomic_id"]
        self.workflow.set_device(create_device("aws", self.device_arn))

        self.workflow.compile(self.workflow.problem)

        self.workflow.header["atomic_id"] = self.atomic_uuid

    def run_workflow(self):
        """
        Function to execute the workflow on the Jobs instance

        .. note::
            Note tha this function is executed within the AWS job-docker
        """

        try:
            self.workflow.optimize()
            self.completed = True
            print("Job completed!!!!")
        except Exception as e:
            print("An Exception has occurred when to run the workflow: {}".format(e))
            return False



import numpy as np
from typing import Union, List, Tuple, Optional

# IBM Qiskit imports
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.providers.aer import AerSimulator
from qiskit.providers.aer.noise import NoiseModel
from qiskit.opflow.primitive_ops import PauliSumOp
from qiskit.quantum_info import Statevector
from qiskit.circuit import Parameter

from .gates_qiskit import QiskitGateApplicator
from openqaoa.backends.basebackend import (
    QAOABaseBackendParametric,
    QAOABaseBackendShotBased,
    QAOABaseBackendStatevector,
)
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from openqaoa.utilities import (
    flip_counts,
    generate_uuid,
    round_value,
)
from openqaoa.backends.cost_function import cost_function
from openqaoa.qaoa_components.ansatz_constructor import (
    RXGateMap,
    RYGateMap,
    RZGateMap,
    RXXGateMap,
    RYYGateMap,
    RZZGateMap,
    RZXGateMap,
)


class QAOAQiskitBackendShotBasedSimulator(
    QAOABaseBackendShotBased, QAOABaseBackendParametric
):
    """
    Local Shot-based simulators offered by Qiskit

    Parameters
    ----------
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.

    n_shots: `int`
        The number of shots to be taken for each circuit.

    seed_simulator: `int`
        Pseudorandom list of numbers of a seed

    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.

    append_state: `QuantumCircuit`
        The state appended to the circuit.

    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.

    cvar_alpha: `float`
        The value of alpha for the CVaR cost function.

    qiskit_simulation_method: `str`
        The method to be used for the simulation.

    noise_model: `NoiseModel`
        The Qiskit noise model to be used for the simulation.
    """

    QISKIT_GATEMAP_LIBRARY = [
        RXGateMap,
        RYGateMap,
        RZGateMap,
        RXXGateMap,
        RYYGateMap,
        RZZGateMap,
        RZXGateMap,
    ]

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        n_shots: int,
        prepend_state: Optional[QuantumCircuit],
        append_state: Optional[QuantumCircuit],
        init_hadamard: bool,
        cvar_alpha: float,
        qiskit_simulation_method: str = "automatic",
        seed_simulator: Optional[int] = None,
        noise_model: Optional[NoiseModel] = None,
    ):
        QAOABaseBackendShotBased.__init__(
            self,
            qaoa_descriptor,
            n_shots,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )
        self.gate_applicator = QiskitGateApplicator()
        self.qureg = QuantumRegister(self.n_qubits)

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), (
                "Cannot attach a bigger circuit " "to the QAOA routine"
            )
        # options = {"seed_simulator":1}
        self.backend_simulator = AerSimulator(
            method=qiskit_simulation_method.lower(),
            noise_model=noise_model,
            seed_simulator=seed_simulator,
        )
        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """
        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        circuit_with_angles = self.parametric_circuit.bind_parameters(memory_map)

        if self.append_state:
            circuit_with_angles = circuit_with_angles.compose(self.append_state)
        circuit_with_angles.measure_all()

        return circuit_with_angles

    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(
            self.qureg
        )  # consider changing this too with my new function

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.qureg)

        self.qiskit_parameter_list = []
        for each_gate in self.abstract_circuit:
            # if gate is of type mixer or cost gate, assign parameter to it
            if each_gate.gate_label.type.value in ["MIXER", "COST"]:
                angle_param = Parameter(each_gate.gate_label.__repr__())
                self.qiskit_parameter_list.append(angle_param)
                each_gate.angle_value = angle_param
            if (
                type(each_gate)
                in QAOAQiskitBackendShotBasedSimulator.QISKIT_GATEMAP_LIBRARY
            ):
                decomposition = each_gate.decomposition("trivial")
            else:
                decomposition = each_gate.decomposition("standard")
            # Create Circuit
            for each_tuple in decomposition:
                gate = each_tuple[0](self.gate_applicator, *each_tuple[1])
                gate.apply_gate(parametric_circuit)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Returns the counts of the final QAOA circuit after binding angles from variational parameters.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing variable parameters.
        n_shots: `int`
            The number of times to run the circuit. If None, n_shots is set to the default: self.n_shots

        Returns
        -------
        counts: `dict`
            The counts of the final QAOA circuit after binding angles from variational parameters.
        """
        # generate a job id for the wavefunction evaluation
        self.job_id = generate_uuid()

        # set the number of shots, if not specified take the default
        n_shots = self.n_shots if n_shots == None else n_shots

        qaoa_circuit = self.qaoa_circuit(params)
        counts = (
            self.backend_simulator.run(qaoa_circuit, shots=n_shots)
            .result()
            .get_counts()
        )

        final_counts = flip_counts(counts)
        self.measurement_outcomes = final_counts
        return final_counts

    def circuit_to_qasm(self):
        """
        A method to convert the QAOA circuit to QASM.
        """
        raise NotImplementedError()

    #         qasm_circuit = self.parametric_circuit.qasm()
    #         return qasm_circuit

    def reset_circuit(self):
        raise NotImplementedError()


class QAOAQiskitBackendStatevecSimulator(
    QAOABaseBackendStatevector, QAOABaseBackendParametric
):
    """
    Local Statevector-based simulators offered by Qiskit

    Parameters
    ----------
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.

    n_shots: `int`
        The number of shots to be taken for each circuit.

    prepend_state: `np.ndarray` or `QuantumCircuit`
        The state prepended to the circuit.

    append_state: `QuantumCircuit or np.ndarray`
        The state appended to the circuit.

    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.

    cvar_alpha: `float`
        The value of alpha for the CVaR cost function.
    """

    QISKIT_GATEMAP_LIBRARY = [
        RXGateMap,
        RYGateMap,
        RZGateMap,
        RXXGateMap,
        RYYGateMap,
        RZZGateMap,
        RZXGateMap,
    ]

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        prepend_state: Optional[Union[np.ndarray, QuantumCircuit]],
        append_state: Optional[Union[np.ndarray, QuantumCircuit]],
        init_hadamard: bool,
        cvar_alpha: float = 1,
    ):
        QAOABaseBackendStatevector.__init__(
            self,
            qaoa_descriptor,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )

        assert (
            cvar_alpha == 1
        ), "Please use the shot-based simulator for simulations with cvar_alpha < 1"

        self.qureg = QuantumRegister(self.n_qubits)
        self.gate_applicator = QiskitGateApplicator()

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), (
                "Cannot attach a bigger circuit " "to the QAOA routine"
            )

        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit
        self.qiskit_cost_hamil = self.qiskit_cost_hamiltonian
        self.qiskit_cost_hamil_sq = self.qiskit_cost_hamil**2

    @property
    def qiskit_cost_hamiltonian(self):
        """
        The qiskit cost hamiltonian for the QAOA circuit represented
        as a `PauliSumOp` object.
        """
        cost_hamil = self.cost_hamiltonian
        n_qubits = cost_hamil.n_qubits
        pauli_strings_list = ["I" * n_qubits] * len(cost_hamil.terms)
        for i, pauli_op in enumerate(cost_hamil.terms):
            pauli_term = list(pauli_strings_list[i])
            for pauli, qubit in zip(pauli_op.pauli_str, pauli_op.qubit_indices):
                pauli_term[qubit] = pauli
            # reverse pauli_strings because qiskit supports little endian notation
            pauli_strings_list[i] = "".join(str(term) for term in pauli_term)[::-1]

        pauli_strings_list.append("I" * n_qubits)
        pauli_coeffs = cost_hamil.coeffs

        qiskit_pauli_op = [
            [pauli_strings, coeff]
            for pauli_strings, coeff in zip(pauli_strings_list, pauli_coeffs)
        ]
        qiskit_pauli_op.append(["I" * n_qubits, cost_hamil.constant])
        qiskit_cost_hamil = PauliSumOp.from_list(qiskit_pauli_op)
        return qiskit_cost_hamil

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """
        # generate a job id for the wavefunction evaluation
        self.job_id = generate_uuid()

        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        circuit_with_angles = self.parametric_circuit.bind_parameters(memory_map)
        return circuit_with_angles

    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
            params:
                Object of type QAOAVariationalBaseParams
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(self.qureg)

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.qureg)

        self.qiskit_parameter_list = []
        for each_gate in self.abstract_circuit:
            # if gate is of type mixer or cost gate, assign parameter to it
            if each_gate.gate_label.type.value in ["MIXER", "COST"]:
                angle_param = Parameter(each_gate.gate_label.__repr__())
                self.qiskit_parameter_list.append(angle_param)
                each_gate.angle_value = angle_param
            if (
                type(each_gate)
                in QAOAQiskitBackendStatevecSimulator.QISKIT_GATEMAP_LIBRARY
            ):
                decomposition = each_gate.decomposition("trivial")
            else:
                decomposition = each_gate.decomposition("standard")
            # Create Circuit
            for each_tuple in decomposition:
                gate = each_tuple[0](self.gate_applicator, *each_tuple[1])
                gate.apply_gate(parametric_circuit)

        if self.append_state:
            parametric_circuit = parametric_circuit.compose(self.append_state)

        return parametric_circuit

    def wavefunction(
        self, params: QAOAVariationalBaseParams
    ) -> Union[List[complex], np.ndarray]:
        """
        Get the wavefunction of the state produced by the parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        wf: `List[complex]` or `np.ndarray[complex]`
            A list of the wavefunction amplitudes.
        """
        ckt = self.qaoa_circuit(params)
        wf = Statevector(ckt).data
        self.measurement_outcomes = wf
        return wf

    @round_value
    def expectation(self, params: QAOAVariationalBaseParams) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.

        Returns
        -------
        `float`
            expectation value of cost operator wrt to quantum state produced by QAOA circuit
        """
        ckt = self.qaoa_circuit(params)
        output_wf = Statevector(ckt)
        self.measurement_outcomes = output_wf.data
        cost = np.real(output_wf.expectation_value(self.qiskit_cost_hamil))
        return cost

    @round_value
    def expectation_w_uncertainty(
        self, params: QAOAVariationalBaseParams
    ) -> Tuple[float, float]:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian and its uncertainty

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.

        Returns
        -------
        `Tuple[float]`
            expectation value and its uncertainty of cost operator wrt
            to quantum state produced by QAOA circuit.
        """
        ckt = self.qaoa_circuit(params)
        output_wf = Statevector(ckt)
        self.measurement_outcomes = output_wf.data
        cost = np.real(Statevector(ckt).expectation_value(self.qiskit_cost_hamil))
        cost_sq = np.real(Statevector(ckt).expectation_value(self.qiskit_cost_hamil_sq))

        uncertainty = np.sqrt(cost_sq - cost**2)
        return (cost, uncertainty)

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation

        TODO: Check if only the first instruction is needed. Two might
              be redundant
        """
        raise NotImplementedError()

    def circuit_to_qasm(self):
        """ """
        raise NotImplementedError()



import time
import unittest
from unittest.mock import Mock
import json
import numpy as np
import pytest
import subprocess

from qiskit import QuantumCircuit
from qiskit.quantum_info import Operator
from qiskit.tools.monitor import job_monitor

from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    PauliOp,
    Hamiltonian,
    QAOADescriptor,
    QAOAVariationalStandardParams,
)
from openqaoa_qiskit.backends import (
    DeviceQiskit,
    QAOAQiskitQPUBackend,
    QAOAQiskitBackendStatevecSimulator,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.problems import NumberPartition
from openqaoa import QAOA, create_device


class TestingQAOAQiskitQPUBackend(unittest.TestCase):

    """This Object tests the QAOA Qiskit QPU Backend objects, which is tasked with the
    creation and execution of a QAOA circuit for the selected QPU provider and
    backend.

    IBMQ Account has to be saved locally to run these tests.
    """

    @pytest.mark.api
    def setUp(self):
        """
        Define the credentials
        """
        self.HUB = "ibm-q"
        self.GROUP = "open"
        self.PROJECT = "main"

    @pytest.mark.api
    def test_circuit_angle_assignment_qpu_backend(self):
        """
        A tests that checks if the circuit created by the Qiskit Backend
        has the appropriate angles assigned before the circuit is executed.
        Checks the circuit created on both IBM QPU Backends.
        """

        nqubits = 3
        p = 2
        weights = [1, 1, 1]
        gammas = [0, 1 / 8 * np.pi]
        betas = [1 / 2 * np.pi, 3 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_device = DeviceQiskit(
            "ibmq_qasm_simulator", self.HUB, self.GROUP, self.PROJECT
        )

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, None, None, False
        )
        qpu_circuit = qiskit_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[0], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[1], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[1], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[1], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[1], 0)
        main_circuit.rx(-2 * betas[1], 1)
        main_circuit.rx(-2 * betas[1], 2)

        qpu_circuit.remove_final_measurements(inplace=True)
        qpu_circuit_operator = Operator(qpu_circuit)
        main_circuit_operator = Operator(main_circuit)

        assert qpu_circuit_operator.equiv(main_circuit_operator)

    @pytest.mark.api
    def test_circuit_angle_assignment_qpu_backend_w_hadamard(self):
        """
        Checks for consistent if init_hadamard is set to True.
        """

        nqubits = 3
        p = 2
        weights = [1, 1, 1]
        gammas = [0, 1 / 8 * np.pi]
        betas = [1 / 2 * np.pi, 3 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_device = DeviceQiskit(
            "ibmq_qasm_simulator", self.HUB, self.GROUP, self.PROJECT
        )

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, None, None, True
        )
        qpu_circuit = qiskit_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.h([0, 1, 2])
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[0], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[1], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[1], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[1], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[1], 0)
        main_circuit.rx(-2 * betas[1], 1)
        main_circuit.rx(-2 * betas[1], 2)

        qpu_circuit.remove_final_measurements(inplace=True)
        qpu_circuit_operator = Operator(qpu_circuit)
        main_circuit_operator = Operator(main_circuit)

        assert qpu_circuit_operator.equiv(main_circuit_operator)

    @pytest.mark.api
    def test_prepend_circuit(self):
        """
        Checks if prepended circuit has been prepended correctly.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        # Prepended Circuit
        prepend_circuit = QuantumCircuit(3)
        prepend_circuit.x([0, 1, 2])

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_device = DeviceQiskit(
            "ibmq_qasm_simulator", self.HUB, self.GROUP, self.PROJECT
        )

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, prepend_circuit, None, True
        )
        qpu_circuit = qiskit_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.x([0, 1, 2])
        main_circuit.h([0, 1, 2])
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[0], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)

        qpu_circuit.remove_final_measurements(inplace=True)
        qpu_circuit_operator = Operator(qpu_circuit)
        main_circuit_operator = Operator(main_circuit)

        assert qpu_circuit_operator.equiv(main_circuit_operator)

    @pytest.mark.api
    def test_append_circuit(self):
        """
        Checks if appended circuit is appropriately appended to the back of the
        QAOA Circuit.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        # Appended Circuit
        append_circuit = QuantumCircuit(3)
        append_circuit.x([0, 1, 2])

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_device = DeviceQiskit(
            "ibmq_qasm_simulator", self.HUB, self.GROUP, self.PROJECT
        )

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, None, append_circuit, True
        )
        qpu_circuit = qiskit_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.h([0, 1, 2])
        main_circuit.cx(0, 1)
        main_circuit.rz(2 * gammas[0], 1)
        main_circuit.cx(0, 1)
        main_circuit.cx(1, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(1, 2)
        main_circuit.cx(0, 2)
        main_circuit.rz(2 * gammas[0], 2)
        main_circuit.cx(0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)
        main_circuit.x([0, 1, 2])

        qpu_circuit.remove_final_measurements(inplace=True)
        qpu_circuit_operator = Operator(qpu_circuit)
        main_circuit_operator = Operator(main_circuit)

        assert qpu_circuit_operator.equiv(main_circuit_operator)

    @pytest.mark.api
    def test_expectations_in_init(self):
        """
        Testing the Exceptions in the init function of the QiskitQPUShotBasedBackend
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)

        # Check the creation of the varitional parms
        _ = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        # We mock the potential Exception that could occur in the Device class
        qiskit_device = DeviceQiskit("", "", "", "")
        qiskit_device._check_provider_connection = Mock(return_value=False)

        try:
            QAOAQiskitQPUBackend(
                qaoa_descriptor, qiskit_device, shots, None, None, True
            )
        except Exception as e:
            self.assertEqual(str(e), "Error connecting to IBMQ.")

        self.assertRaises(
            Exception,
            QAOAQiskitQPUBackend,
            (qaoa_descriptor, qiskit_device, shots, None, None, True),
        )

        qiskit_device = DeviceQiskit(
            device_name="",
            hub=self.HUB,
            group=self.GROUP,
            project=self.PROJECT,
        )

        try:
            QAOAQiskitQPUBackend(
                qaoa_descriptor, qiskit_device, shots, None, None, True
            )
        except Exception as e:
            self.assertEqual(
                str(e),
                "Connection to IBMQ was made. Error connecting to the specified backend.",
            )

        self.assertRaises(
            Exception,
            QAOAQiskitQPUBackend,
            qaoa_descriptor,
            qiskit_device,
            shots,
            None,
            None,
            True,
        )

    @pytest.mark.sim
    def test_remote_integration_sim_run(self):
        """
        Checks if Remote IBM QASM Simulator is similar/close to Local IBM
        Statevector Simulator.
        This test also serves as an integration test for the IBMQPU Backend.

        This test takes a long time to complete.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [[0], [1 / 8 * np.pi], [0], [1 / 8 * np.pi]]
        betas = [[0], [0], [1 / 8 * np.pi], [1 / 8 * np.pi]]
        shots = 10000

        for i in range(4):
            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights,
                1,
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas[i], gammas[i]
            )

            qiskit_device = DeviceQiskit(
                "ibmq_qasm_simulator", self.HUB, self.GROUP, self.PROJECT
            )

            qiskit_backend = QAOAQiskitQPUBackend(
                qaoa_descriptor, qiskit_device, shots, None, None, False
            )
            qiskit_expectation = qiskit_backend.expectation(variate_params)

            qiskit_statevec_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, False
            )
            qiskit_statevec_expectation = qiskit_statevec_backend.expectation(
                variate_params
            )

            acceptable_delta = 0.05 * qiskit_statevec_expectation
            self.assertAlmostEqual(
                qiskit_expectation, qiskit_statevec_expectation, delta=acceptable_delta
            )

    @pytest.mark.api
    def test_remote_qubit_overflow(self):
        """
        If the user creates a circuit that is larger than the maximum circuit size
        that is supported by the QPU. An Exception should be raised with the
        appropriate error message alerting the user to the error.
        """

        shots = 100

        set_of_numbers = np.random.randint(1, 10, 6).tolist()
        qubo = NumberPartition(set_of_numbers).qubo

        mixer_hamil = X_mixer_hamiltonian(n_qubits=6)
        qaoa_descriptor = QAOADescriptor(qubo.hamiltonian, mixer_hamil, p=1)

        # Check the creation of the varitional parms
        _ = create_qaoa_variational_params(qaoa_descriptor, "standard", "rand")

        qiskit_device = DeviceQiskit("ibmq_manila", self.HUB, self.GROUP, self.PROJECT)

        try:
            QAOAQiskitQPUBackend(
                qaoa_descriptor, qiskit_device, shots, None, None, True
            )
        except Exception as e:
            self.assertEqual(
                str(e),
                "There are lesser qubits on the device than the number of qubits required for the circuit.",
            )

    @pytest.mark.qpu
    def test_integration_on_emulator(self):
        """
        Test Emulated QPU Workflow. Checks if the expectation value is returned
        after the circuit run.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [[1 / 8 * np.pi]]
        betas = [[1 / 8 * np.pi]]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)
        qiskit_device = DeviceQiskit(
            "ibm_perth",
            self.HUB,
            self.GROUP,
            self.PROJECT,
            as_emulator=True,
        )

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, None, None, False
        )
        qiskit_expectation = qiskit_backend.expectation(variate_params)

        self.assertEqual(type(qiskit_expectation.item()), float)

    @pytest.mark.qpu
    def test_remote_integration_qpu_run(self):
        """
        Test Actual QPU Workflow. Checks if the expectation value is returned
        after the circuit run.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [[1 / 8 * np.pi]]
        betas = [[1 / 8 * np.pi]]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)
        qiskit_device = DeviceQiskit("ibm_perth", self.HUB, self.GROUP, self.PROJECT)

        qiskit_backend = QAOAQiskitQPUBackend(
            qaoa_descriptor, qiskit_device, shots, None, None, False
        )
        circuit = qiskit_backend.qaoa_circuit(variate_params)
        job = qiskit_backend.backend_qpu.run(circuit, shots=qiskit_backend.n_shots)

        # check if the cirucit is validated by IBMQ servers when submitted for execution
        # check the status of the job and keep retrying until its completed or queued
        while job.status().name not in ["DONE", "CANCELLED", "ERROR"]:
            # if the job is queued, cancel the job
            if job.status().name == "QUEUED":
                job.cancel()
            else:
                time.sleep(1)

        assert job.status().name in ["DONE", "CANCELLED"]


if __name__ == "__main__":
    unittest.main()



import networkx as nx

from ..utilities import check_kwargs
from .problem import Problem
from .qubo import QUBO


class MaximumCut(Problem):
    """
    Creates an instance of the Maximum Cut problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.

    Returns
    -------
        An instance of the Maximum Cut problem.
    """

    __name__ = "maximum_cut"

    DEFAULT_EDGE_WEIGHT = 1.0

    def __init__(self, G):
        self.G = G

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Maximum Cut problem, whose graph is
        random following the Erdos-Renyi model.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
            A random instance of the Maximum Cut problem.
        """
        n_nodes, edge_probability = check_kwargs(
            ["n_nodes", "edge_probability"], [None, None], **kwargs
        )
        seed = kwargs.get("seed", None)

        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )
        return MaximumCut(G)

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        # Iterate over edges (with weight) and store accordingly
        terms = []
        weights = []

        for u, v, edge_weight in self.G.edges(data="weight"):
            terms.append([u, v])

            # We expect the edge weight to be given in the attribute called
            # "weight". If it is None, assume a weight of 1.0
            weights.append(
                edge_weight if edge_weight else MaximumCut.DEFAULT_EDGE_WEIGHT
            )

        return QUBO(self.G.number_of_nodes(), terms, weights, self.problem_instance)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Overview
--------

The PennyLane NumPy subpackage provides a differentiable wrapper around NumPy, that enables
backpropagation through standard NumPy code.

This version of NumPy **must** be used when using PennyLane with the :doc:`Autograd interface
</introduction/interfaces/numpy>`:

>>> from pennylane import numpy as np

.. note::

    If using other interfaces, such as :doc:`TensorFlow </introduction/interfaces/tf>` :doc:`PyTorch
    </introduction/interfaces/torch>`, or :doc:`JAX </introduction/interfaces/jax>`, then the
    PennyLane-provided NumPy should not be used; instead, simply use the standard NumPy import.

This package is a wrapper around ``autograd.numpy``; for details on all available functions,
please refer to the `Autograd
docs <https://github.com/HIPS/autograd/blob/master/docs/tutorial.md>`__.

PennyLane additionally extends Autograd with the following classes,
errors, and functions:

.. autosummary::
    :toctree: api
    :nosignatures:
    :template: autosummary/class_no_inherited.rst

    ~wrap_arrays
    ~extract_tensors
    ~tensor_wrapper
    ~tensor
    ~NonDifferentiableError

Caveats
-------

This package is a wrapper around ``autograd.numpy``, and therefore comes with several caveats
inherited from Autograd:

**Do not use:**

- Assignment to arrays, such as ``A[0, 0] = x``.

..

- Implicit casting of lists to arrays, for example ``A = np.sum([x, y])``.
  Make sure to explicitly cast to a NumPy array first, i.e.,
  ``A = np.sum(np.array([x, y]))`` instead.

..

- ``A.dot(B)`` notation. Use ``np.dot(A, B)`` or ``A @ B`` instead.

..

- In-place operations such as ``a += b``. Use ``a = a + b`` instead.

..

- Some ``isinstance`` checks, like ``isinstance(x, np.ndarray)`` or ``isinstance(x, tuple)``,
  without first doing ``from autograd.builtins import isinstance, tuple``.

For more details, please consult the `Autograd
docs <https://github.com/HIPS/autograd/blob/master/docs/tutorial.md>`__.

"""
# pylint: disable=wrong-import-position,wildcard-import,undefined-variable
from autograd import numpy as _np
from autograd.numpy import *

from .wrapper import wrap_arrays, extract_tensors, tensor_wrapper

wrap_arrays(_np.__dict__, globals())

# Delete the unwrapped fft, linalg, random modules
# so that we can re-import our wrapped versions.
del fft
del linalg
del random

from . import fft
from . import linalg
from . import random

from .tensor import tensor, NonDifferentiableError, asarray as _asarray

asarray = tensor_wrapper(_asarray)



from collections import defaultdict
from typing import List
import json
import numpy as np
import scipy

from ..qaoa_components import Hamiltonian
from ..utilities import delete_keys_from_dict, convert2serialize


class QUBO(object):
    """
    Creates an instance of Quadratic Unconstrained Binary Optimization (QUBO)
    class, which offers a way to encode optimization problems.

    Parameters
    ----------
    n: int
        The number of variables in the representation.
    terms: List[Tuple[int, ...],List]
        The different terms in the QUBO encoding, indicating the
        different interactions between variables.
    weights: List[float]
        The list of weights (or coefficients) corresponding to each
        interaction defined in `terms`.
    clean_terms_and_weights: bool
        Boolean indicating whether terms and weights can be cleaned


    Returns
    -------
        An instance of the Quadratic Unconstrained Binary Optimization (QUBO) class.
    """

    # Maximum number of terms allowed to enable the cleaning procedure
    TERMS_CLEANING_LIMIT = 5000

    def __init__(
        self,
        n,
        terms,
        weights,
        problem_instance: dict = {"problem_type": "generic_qubo"},
        clean_terms_and_weights=False,
    ):
        # check-type for terms and weights
        if not isinstance(terms, list) and not isinstance(terms, tuple):
            raise TypeError(
                "The input parameter terms must be of type of list or tuple"
            )

        if not isinstance(weights, list) and not isinstance(weights, tuple):
            raise TypeError(
                "The input parameter weights must be of type of list or tuple"
            )

        for each_entry in weights:
            if not isinstance(each_entry, float) and not isinstance(each_entry, int):
                raise TypeError(
                    "The elements in weights list must be of type float or int."
                )

        terms = list(terms)
        weights = list(weights)

        # Check that terms and weights have matching lengths
        if len(terms) != len(weights):
            raise ValueError("The number of terms and number of weights do not match")

        constant = 0
        try:
            constant_index = [i for i, term in enumerate(terms) if len(term) == 0][0]
            constant = weights.pop(constant_index)
            terms.pop(constant_index)
        except:
            pass

        # If the user wants to clean the terms and weights or if the number of
        # terms is not too big, we go through the cleaning process
        if clean_terms_and_weights or len(terms) <= QUBO.TERMS_CLEANING_LIMIT:
            self.terms, self.weights = QUBO.clean_terms_and_weights(terms, weights)
        else:
            self.terms, self.weights = terms, weights

        self.constant = constant
        self.n = n

        # attribute to store the problem instance, it will be checked
        # if it is json serializable in the __setattr__ method
        self.problem_instance = problem_instance

        # Initialize the metadata dictionary
        self.metadata = {}

    def __iter__(self):
        for key, value in self.__dict__.items():
            # remove "_" from the beginning of the key if it exists
            yield (key[1:] if key.startswith("_") else key, value)

    def __setattr__(self, __name, __value):
        # check if problem_instance is json serializable, also check if
        # metadata is json serializable
        if __name == "problem_instance" or __name == "metadata":
            try:
                _ = json.dumps(__value)
            except Exception as e:
                raise e

        super().__setattr__(__name, __value)

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, input_n):
        if not isinstance(input_n, int):
            raise TypeError("The input parameter, n, has to be of type int")

        if input_n <= 0:
            raise TypeError(
                "The input parameter, n, must be a positive integer greater than 0"
            )

        self._n = input_n

    def set_metadata(self, metadata: dict = {}):
        """


        Parameters
        ----------
        metadata: dict
            The metadata of the problem. All keys and values will
            be stored in the metadata dictionary.
        """

        # update the metadata (it will be checked if it is json
        # serializable in the __setattr__ method)
        self.metadata = {**self.metadata, **metadata}

    def asdict(self, exclude_keys: List[str] = []):
        """
        Returns a dictionary containing the serialization of the class.

        Parameters
        ----------
        exclude_keys: List[str]


        Returns
        -------
            A dictionary containing the serialization of the class.
        """

        if exclude_keys == []:
            return convert2serialize(dict(self))
        else:
            return delete_keys_from_dict(
                obj=convert2serialize(dict(self)), keys_to_delete=exclude_keys
            )

    @staticmethod
    def from_dict(dict: dict, clean_terms_and_weights=False):
        """


        Parameters
        ----------
        dict: dict
            The dictionary containing the serialization of the QUBO object.
        clean_terms_and_weights: bool


        Returns
        -------
            A QUBO object.
        """

        # make a copy of the dictionary to avoid modifying the input
        dict = dict.copy()

        # extract the metadata
        metadata = dict.pop("metadata", {})

        # make a copy of the terms and weights to avoid modifying the input
        dict["terms"] = dict["terms"].copy()
        dict["weights"] = dict["weights"].copy()

        # add the constant term
        dict["terms"].append([])
        dict["weights"].append(dict.pop("constant", 0))

        # create the QUBO object
        qubo = QUBO(**dict, clean_terms_and_weights=clean_terms_and_weights)

        # add the metadata
        qubo.metadata = metadata.copy()

        # return the QUBO object
        return qubo

    @staticmethod
    def clean_terms_and_weights(terms, weights):
        """Goes through the terms and weights and group them when possible"""
        # List to record the terms as sets
        unique_terms = []

        # Will record the weight for the unique terms (note that since Sets are
        # unhashable in Python, we use a dict with integers for the keys, that
        # are mapped with the corresponding indices of terms from unique_terms)
        new_weights_for_terms = defaultdict(float)

        # We do one pass over terms and weights
        for term, weight in zip(terms, weights):
            # Convert the term to a set
            term_set = set(term)

            # If this term is not yet recorded, we add it to the list of unique
            # terms and we use that it is the last element to find its index
            if term_set not in unique_terms:
                unique_terms.append(term_set)
                term_index = len(unique_terms) - 1

            # Else if the term is alreaddy recorded, we just need to retrieve
            # its index in the unique_terms list
            else:
                term_index = unique_terms.index(term_set)

            # Update the weight in the dictionary using the retrieved index
            new_weights_for_terms[term_index] += weight

        # Return terms and weights, making sure to convert the terms back to lists
        return (
            [list(term) for term in unique_terms],
            list(new_weights_for_terms.values()),
        )

    @staticmethod
    def random_instance(n, density=0.5, format_m="coo", max_abs_value=100):
        # Generate a random matrix (elements in [0, 1]) of type sparse
        random_matrix = scipy.sparse.rand(n, n, density=density, format=format_m)

        # Retrieve the indices of non-zero elements of the matrix as list of tuples
        terms = np.transpose(random_matrix.nonzero())

        # Get the matrix entries in a list, but scale the elements and
        # make them centered at 0 by subtracting 0.5
        weights = max_abs_value * (random_matrix.data - 0.5)

        # Return the terms and weights, taking care of converting to the correct types
        return QUBO(n, [list(map(int, i)) for i in terms], [float(i) for i in weights])

    @staticmethod
    def convert_qubo_to_ising(n, qubo_terms, qubo_weights):
        """Convert QUBO terms and weights to their Ising equivalent"""
        ising_terms, ising_weights = [], []
        constant_term = 0
        linear_terms = np.zeros(n)

        # Process the given terms and weights
        for term, weight in zip(qubo_terms, qubo_weights):
            if len(term) == 2:
                u, v = term

                if u != v:
                    ising_terms.append([u, v])
                    ising_weights.append(weight / 4)
                else:
                    constant_term += weight / 4

                linear_terms[term[0]] -= weight / 4
                linear_terms[term[1]] -= weight / 4
                constant_term += weight / 4
            elif len(term) == 1:
                linear_terms[term[0]] -= weight / 2
                constant_term += weight / 2
            else:
                constant_term += weight

        for variable, linear_term in enumerate(linear_terms):
            ising_terms.append([variable])
            ising_weights.append(linear_term)

        ising_terms.append([])
        ising_weights.append(constant_term)
        return ising_terms, ising_weights

    @property
    def hamiltonian(self):
        """
        Returns the Hamiltonian of the problem.
        """
        return Hamiltonian.classical_hamiltonian(
            self.terms, self.weights, self.constant
        )



from typing import List, Union
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

from .variational_baseparams import QAOAVariationalBaseParams
from ..ansatz_constructor import QAOADescriptor
from ..ansatz_constructor.baseparams import shapedArray


class QAOAVariationalAnnealingParams(QAOAVariationalBaseParams):
    """
    QAOA parameters that implement a state preparation circuit of the form

    .. math::

        U = e^{-i (1-s(t_p)) H_M \Delta t} e^{-i s(t_p) H_C \Delta t} \cdots e^{-i(1-s(t_1)H_M \Delta t} e^{-i s(t_1) H_C \Delta t}

    where the :math:`s(t_i) =: s_i` are the variable parameters and
    :math:`\\Delta t= \\frac{T}{N}`.
    So the schedule is specified by specifying s(t) at evenly spaced timelayers.

    Parameters
    ----------
    qaoa_descriptor:
        QAOADescriptor object containing circuit instructions
    total_annealing_time: float
        Total annealing time for the schedule
    schedule: list
        List specifying the annealing schedule

    Attributes
    ----------
    schedule: np.array
        An 1D array holding the values of the schedule function at each timestep.
    total_annealing_time: float
        Total annealing time for the schedule
    dt: float
        annealing time step
    mixer_time: np.array
        schedule for mixer application
    cost_time: np.array
        schedule for cost application
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        total_annealing_time: float,
        schedule: List[Union[float, int]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)
        assert (
            total_annealing_time is not None
        ), f"Please specify total_annealing_time to use {type(self).__name__}"
        self.total_annealing_time = total_annealing_time
        self.schedule = np.array(schedule)
        self.schedule = np.array(schedule)

        self.dt = self.total_annealing_time / self.p
        self.mixer_time = (1 - self.schedule) * self.dt
        self.cost_time = self.schedule * self.dt

    def __repr__(self):
        string = "Annealing Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Variational Parameters:\n"
        string += "\tschedule: " + str(self.schedule)
        return string

    def __len__(self):
        return self.p

    @shapedArray
    def schedule(self):
        return self.p

    @property
    def mixer_1q_angles(self):
        return 2 * np.outer(self.mixer_time, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self) -> np.ndarray:
        return 2 * np.outer(self.mixer_time, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2 * np.outer(self.cost_time, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2 * np.outer(self.cost_time, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        if len(new_values) != self.p:
            raise RuntimeWarning("the new times should have length p")
        self.schedule = np.array(new_values)
        # udpate mixer_time and cost_time too
        self.mixer_time = (1 - self.schedule) * self.dt
        self.cost_time = self.schedule * self.dt

    def raw(self):
        return self.schedule

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls,
        qaoa_descriptor: QAOADescriptor,
        total_annealing_time: float = None,
        time: float = None,
    ):
        """
        Returns
        -------
        AnnealingParams :
            An ``AnnealingParams`` object corresponding to
            a linear ramp schedule.
        """
        p = qaoa_descriptor.p
        total_annealing_time = (
            0.7 * p if total_annealing_time is None else total_annealing_time
        )
        schedule = np.linspace(0.5 / p, 1 - 0.5 / p, p)

        # wrap it all nicely in a qaoa_parameters object
        # params = cls((register, terms, weights, p, time), (schedule))
        params = cls(qaoa_descriptor, total_annealing_time, schedule)
        return params

    @classmethod
    def random(
        cls,
        qaoa_descriptor: QAOADescriptor,
        total_annealing_time: float,
        seed: int = None,
    ):
        """
        Returns
        -------
        AnnealingParams
            randomly initialised AnnealingParams object
        """
        if seed is not None:
            np.random.seed(seed)
        schedule = np.random.uniform(0, 1, qaoa_descriptor.p)

        return cls(qaoa_descriptor, total_annealing_time, schedule)

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor, total_annealing_time: float):
        schedule = np.empty((qaoa_descriptor.p))
        return cls(qaoa_descriptor, total_annealing_time, schedule)

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(self.schedule, marker="s", **kwargs)
        ax.set_xlabel("p", fontsize=14)
        ax.set_ylabel("s(t)", fontsize=14)
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))



import time
from typing import Optional, List
import warnings

# IBM Qiskit imports
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, transpile
from qiskit_ibm_provider.job.exceptions import (
    IBMJobApiError,
    IBMJobInvalidStateError,
    IBMJobFailureError,
    IBMJobTimeoutError,
)
from qiskit.circuit import Parameter

from .devices import DeviceQiskit
from .gates_qiskit import QiskitGateApplicator
from openqaoa.backends.basebackend import (
    QAOABaseBackendShotBased,
    QAOABaseBackendCloud,
    QAOABaseBackendParametric,
)
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from openqaoa.utilities import flip_counts


class QAOAQiskitQPUBackend(
    QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased
):
    """
    A QAOA simulator as well as for real QPU using qiskit as the backend

    Parameters
    ----------
    device: `DeviceQiskit`
        An object of the class ``DeviceQiskit`` which contains the credentials
        for accessing the QPU via cloud and the name of the device.
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.
    append_state: `QuantumCircuit`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.
    cvar_alpha: `float`
        The value of alpha for the CVaR method.
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        device: DeviceQiskit,
        n_shots: int,
        prepend_state: Optional[QuantumCircuit],
        append_state: Optional[QuantumCircuit],
        init_hadamard: bool,
        initial_qubit_mapping: Optional[List[int]] = None,
        qiskit_optimization_level: int = 1,
        cvar_alpha: float = 1,
    ):
        QAOABaseBackendShotBased.__init__(
            self,
            qaoa_descriptor,
            n_shots,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )
        QAOABaseBackendCloud.__init__(self, device)

        self.qureg = QuantumRegister(self.n_qubits)
        self.problem_reg = self.qureg[0 : self.problem_qubits]
        self.creg = ClassicalRegister(len(self.problem_reg))

        if qiskit_optimization_level in [0, 1, 2, 3]:
            self.qiskit_optimziation_level = qiskit_optimization_level
        else:
            raise ValueError(
                f"qiskit_optimization_level cannot be {qiskit_optimization_level}. Choose between 0 to 3"
            )
        self.gate_applicator = QiskitGateApplicator()

        if self.initial_qubit_mapping is None:
            self.initial_qubit_mapping = (
                initial_qubit_mapping
                if initial_qubit_mapping is not None
                else list(range(self.n_qubits))
            )

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), (
                "Cannot attach a bigger circuit" "to the QAOA routine"
            )

        if self.device.provider_connected and self.device.qpu_connected:
            self.backend_qpu = self.device.backend_device
        elif self.device.provider_connected and self.device.qpu_connected in [
            False,
            None,
        ]:
            raise Exception(
                "Connection to {} was made. Error connecting to the specified backend.".format(
                    self.device.device_location.upper()
                )
            )

        else:
            raise Exception(
                "Error connecting to {}.".format(self.device.device_location.upper())
            )

        if self.device.n_qubits < self.n_qubits:
            raise Exception(
                "There are lesser qubits on the device than the number of qubits required for the circuit."
            )
        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """
        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        circuit_with_angles = self.parametric_circuit.bind_parameters(memory_map)

        if self.append_state:
            circuit_with_angles = circuit_with_angles.compose(self.append_state)

        # only measure the problem qubits
        if self.final_mapping is None:
            circuit_with_angles.measure(self.problem_reg, self.creg)
        else:
            for idx, qubit in enumerate(self.final_mapping[0 : len(self.problem_reg)]):
                cbit = self.creg[idx]
                circuit_with_angles.measure(qubit, cbit)

        if self.qaoa_descriptor.routed is True:
            transpiled_circuit = transpile(
                circuit_with_angles,
                self.backend_qpu,
                initial_layout=self.initial_qubit_mapping,
                optimization_level=self.qiskit_optimziation_level,
                routing_method="none",
            )
        else:
            transpiled_circuit = transpile(
                circuit_with_angles,
                self.backend_qpu,
                initial_layout=self.initial_qubit_mapping,
                optimization_level=self.qiskit_optimziation_level,
            )

        return transpiled_circuit

    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
            params:
                Object of type QAOAVariationalBaseParams
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(self.qureg, self.creg)

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.problem_reg)

        self.qiskit_parameter_list = []
        for each_gate in self.abstract_circuit:
            # if gate is of type mixer or cost gate, assign parameter to it
            if each_gate.gate_label.type.value in ["MIXER", "COST"]:
                angle_param = Parameter(each_gate.gate_label.__repr__())
                self.qiskit_parameter_list.append(angle_param)
                each_gate.angle_value = angle_param
            decomposition = each_gate.decomposition("standard")
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0](self.gate_applicator, *each_tuple[1])
                gate.apply_gate(parametric_circuit)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Execute the circuit and obtain the counts

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: int
            The number of times to run the circuit. If None, n_shots is set to the default: self.n_shots

        Returns
        -------
            A dictionary with the bitstring as the key and the number of counts
            as its value.
        """

        n_shots = self.n_shots if n_shots is None else n_shots

        circuit = self.qaoa_circuit(params)

        job_state = False
        no_of_job_retries = 0
        max_job_retries = 5

        while job_state is False:
            # initial_layout only passed if not azure device
            # if type(self.device).__name__ == "DeviceAzure":
            # job = self.backend_qpu.run(circuit, **input_items)
            job = self.backend_qpu.run(circuit, shots=n_shots)

            api_contact = False
            no_of_api_retries = 0
            max_api_retries = 5

            while api_contact is False:
                try:
                    self.job_id = job.job_id()
                    counts = job.result().get_counts()
                    api_contact = True
                    job_state = True
                except (IBMJobApiError, IBMJobTimeoutError):
                    print("There was an error when trying to contact the IBMQ API.")
                    job_state = True
                    no_of_api_retries += 1
                    time.sleep(5)
                except (IBMJobFailureError, IBMJobInvalidStateError):
                    print("There was an error with the state of the Job in IBMQ.")
                    no_of_job_retries += 1
                    break

                if no_of_api_retries >= max_api_retries:
                    raise ConnectionError(
                        "Number of API Retries exceeded Maximum allowed."
                    )

            if no_of_job_retries >= max_job_retries:
                raise ConnectionError("An Error Occurred with the Job(s) sent to IBMQ.")

        # Expose counts
        final_counts = flip_counts(counts)
        self.measurement_outcomes = final_counts
        return final_counts

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the entire QAOA `QuantumCircuit` object into
        a OpenQASM string
        """
        raise NotImplementedError()
        # qasm_string = self.qaoa_circuit(params).qasm(formatted=True)
        # return qasm_string

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation
        """
        raise NotImplementedError()
